#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

from .. import Unit


class url(Unit):
    """
    Decodes and encodes URL-Encoding, which preserves only alphanumeric
    characters and the symbols `_`, `.`, `-`, `~`, `\\`, and `/`.
    Every other character is escaped by hex-encoding it and prefixing it
    with a percent symbol.
    """

    def interface(self, argp):
        argp.add_argument('-p', '--plus', action='store_true', help='also replace plus signs by spaces')
        return super().interface(argp)

    def process(self, data):
        data = re.sub(
            B'\\%([0-9a-fA-F]{2})',
            lambda m: bytes((int(m.group(1), 16),)),
            data
        )
        if self.args.plus:
            data = data.replace(B'+', B' ')
        return data

    def reverse(self, data):
        if self.args.plus:
            data = data.replace(B' ', B'+')
        data = re.sub(
            B'[^a-zA-Z0-9_.-~\\/]',
            lambda m: B'%%%02X' % ord(m.group(0)),
            data
        )
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from Crypto.Cipher import ARC2

from . import StandardCipherUnit


class rc2(StandardCipherUnit, cipher=ARC2):
    """
    RC2 encryption and decryption.
    """
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import enum
import string
import json

from ... import Unit


class JSONScope(enum.Enum):
    TEXT = B'"'
    LIST = B'['
    DICT = B'{'


class JSONCarver:
    _PRINTABLE_BYTES = set(bytes(string.printable, 'ascii'))
    _MAX_PARSE_DEPTH = 200

    def __init__(self, data, dictonly=False):
        self.data = data
        self.dictonly = dictonly
        self.cursor = 0

    def __iter__(self):
        return self

    def __next__(self):
        while True:
            start = self.data.find(B'{', self.cursor)
            if not self.dictonly:
                lpos = self.data.find(B'[', self.cursor)
                start = lpos if start < 0 else min(lpos, start)
            if start < self.cursor:
                raise StopIteration
            self.cursor = start + 1
            end = self._find_json_end(start)
            if end is None:
                continue
            try:
                if not json.loads(self.data[start:end]):
                    continue
            except json.JSONDecodeError:
                continue
            self.cursor = end + 1
            return self.data[start:end]

    def _find_json_end(self, start):
        token = self.data[start:start + 1]
        scope = [JSONScope(token)]
        cursor = start

        while scope:
            if len(scope) >= self._MAX_PARSE_DEPTH:
                return None
            cursor = cursor + 1
            if cursor >= len(self.data):
                return None
            token = self.data[cursor:cursor + 1]
            if self.data[cursor] not in self._PRINTABLE_BYTES:
                return None
            elif scope[~0] is JSONScope.TEXT:
                if token == B'"' and self.data[cursor - 1:cursor] != B'\\':
                    scope.pop()
                continue
            elif token == B']':
                if scope[~0] is not JSONScope.LIST:
                    return None
                scope.pop()
            elif token == B'}':
                if scope[~0] is not JSONScope.DICT:
                    return None
                scope.pop()
            for t in JSONScope:
                if token == t.value:
                    scope.append(t)
                    continue

        return cursor + 1


class carve_json(Unit):
    """
    Extracts anything from the input data that looks like JSON.
    """

    def interface(self, argp):
        argp.add_argument('-d', '--dictonly', action='store_true',
            help='only extract JSON dictionaries, do not extract lists.')
        return super().interface(argp)

    def process(self, data):
        yield from JSONCarver(data, dictonly=self.args.dictonly)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit
from ...lib.argformats import DelayedBinaryArgument


class sep(Unit):
    """
    Multiple inputs are joined along a specified separator. If any of the input
    `refinery.lib.frame.Chunk`s is currently out of scope, `refinery.sep` turns
    makes them visible by default. This can be prevented by using the `-s` flag.
    """

    def interface(self, argp):
        argp.add_argument('-s', '--scoped', action='store_true',
            help='Prevent sep from automatically turning all input chunks visible.')
        argp.add_argument('separator', type=DelayedBinaryArgument, default=B'\n',
            nargs='?', help='Separator; the default is a line break.')
        return super().interface(argp)

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.separate = False

    def filter(self, inputs):
        it = iter(inputs)
        try:
            data = next(it)
        except StopIteration:
            return
        self.separate = True
        for upcoming in it:
            if not self.args.scoped:
                data.visible = True
            yield data
            data = upcoming
        self.separate = False
        yield data

    def process(self, data):
        yield data
        if self.separate:
            yield self.args.separator



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit
from ...lib.argformats import multibin, number


class pad(Unit):
    """
    Allows padding of the input data. By default, multiple inputs are padded
    to all have length equal to the size of the longest input. Other optional
    size specifications override this behaviour.
    """

    def interface(self, argp):
        size = argp.add_mutually_exclusive_group()
        size.add_argument(
            '-b', '--blocksize',
            type=number,
            default=0,
            metavar='N',
            help='Pad inputs to any even multiple of N.'
        )
        size.add_argument(
            '-a', '--absolute',
            type=number,
            default=0,
            metavar='N',
            help='Pad inputs to be at least N bytes in size.'
        )
        argp.add_argument(
            '-l', '--left',
            action='store_true',
            help='Pad on the left instead of the right.'
        )
        argp.add_argument(
            'padding',
            type=multibin,
            default=B'\0',
            nargs='?',
            help='This custom binary sequence is used (repeatedly, if necessary) '
                 'to pad the input. The default is a zero byte.'
        )
        return super().interface(argp)

    @property
    def relative(self):
        if self.args.blocksize:
            return False
        if self.args.absolute:
            return False
        return True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._maxlen = None

    def _pad(self, data, size):
        missing = (size - len(data))
        if missing <= 0:
            return data
        pad = self.args.padding
        if missing > len(pad):
            pad *= (missing // len(pad))
        if self.args.left:
            return pad[:missing] + data
        else:
            data += pad[:missing]
            return data

    def filter(self, inputs):
        if self.relative:
            self.log_info('padding up to longest input')
            if not isinstance(inputs, list):
                inputs = list(inputs)
            self._maxlen = max(len(d) for d in inputs)
        else:
            self._maxlen = None
        yield from inputs

    def process(self, data):
        if self._maxlen is not None:
            return self._pad(data, self._maxlen)
        if self.args.blocksize:
            q, r = divmod(len(data), self.args.blocksize)
            size = (q + bool(r)) * self.args.blocksize
        else:
            size = self.args.absolute
        return self._pad(data, size)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import defusedxml.minidom

from .. import Unit
from ...lib.decorators import unicoded
from ...lib.argformats import number


class ppxml(Unit):
    """
    Expects XML input data and outputs it in a neatly formatted manner.
    """
    def interface(self, argp):
        argp.add_argument('-i', '--indent', type=number, default=4,
            help='Controls the amount of space characters used for indentation in the output. Default is 4.')
        return super().interface(argp)

    @unicoded
    def process(self, data: str) -> str:
        node = defusedxml.minidom.parseString(data)
        if '<?xml' not in data:
            node = node.childNodes[0]
        return node.toprettyxml(self.args.indent * ' ').strip()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import struct

from . import CipherUnit


class xtea(CipherUnit):
    """
    XTEA encryption and decryption.
    """

    __blocksize__ = 16
    __key_sizes__ = 16

    def encrypt(self, data):
        blocks = iter(self._load32(data))
        key = struct.unpack('IIII', self.key)
        return self._stor64(self._encrypt_block(*bp, *key) for bp in zip(blocks, blocks))

    def decrypt(self, data):
        blocks = iter(self._load32(data))
        key = struct.unpack('IIII', self.key)
        return self._stor64(self._decrypt_block(*z, *key) for z in zip(blocks, blocks))

    @staticmethod
    def _encrypt_block(y, z, k1, k2, k3, k4):
        sum_t = 0
        delta = 0x9E3779B9
        for _ in range(32, 0, -1):
            sum_t = (sum_t + delta) & 0xFFFFFFFF
            y = y + ((z << 4) + k1 ^ z + sum_t ^ (z >> 5) + k2) & 0xFFFFFFFF
            z = z + ((y << 4) + k3 ^ y + sum_t ^ (y >> 5) + k4) & 0xFFFFFFFF
        return y + (z << 0x20)

    @staticmethod
    def _decrypt_block(y, z, k1, k2, k3, k4):
        sum_t = 0xC6EF3720
        delta = 0x9E3779B9
        for _ in range(32, 0, -1):
            z = z - ((y << 4) + k3 ^ y + sum_t ^ (y >> 5) + k4) & 0xFFFFFFFF
            y = y - ((z << 4) + k1 ^ z + sum_t ^ (z >> 5) + k2) & 0xFFFFFFFF
            sum_t = (sum_t - delta) & 0xFFFFFFFF
        return y + (z << 0x20)

    @staticmethod
    def _load32(vector):
        return struct.unpack('I' * (len(vector) // 4), vector)

    @staticmethod
    def _stor64(vector):
        vector = tuple(vector)
        return struct.pack('Q' * len(vector), *vector)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A selection of refinery-specific decorators.
"""
import codecs
import re

from functools import wraps
from typing import Callable, Iterable
from ..units import Unit


def unicoded(method: Callable[[Unit, str], str]) -> Callable[[Unit, bytes], bytes]:
    """
    Can be used to decorate a `refinery.units.Unit.process` routine that takes a
    string argument and also returns one. The resulting routine takes a binary buffer
    as input and attempts to decode it as unicode text. If certain characters cannot
    be decoded, then these ranges are skipped and the decorated routine is called
    once for each string patch that was successfully decoded.
    """
    @wraps(method)
    def method_wrapper(self, data: bytes) -> bytes:
        input_codec = self.codec if any(data[::2]) else 'UTF-16LE'
        partial = re.split(R'([\uDC80-\uDCFF]+)',  # surrogate escape range
            codecs.decode(data, input_codec, errors='surrogateescape'))
        partial[::2] = [method(self, p) if p else '' for p in partial[::2]]
        return codecs.encode(''.join(partial),
            self.codec, errors='surrogateescape')
    return method_wrapper


def linewise(method: Callable[[Unit, str], str]) -> Callable[[Unit, bytes], Iterable[bytes]]:
    """
    Can be used to decorate a `refinery.units.Unit.process` routine that takes a
    string argument and also returns one. The resulting routine expects a default
    encoded string input buffer and calls the decorated routine once for each
    line in the corresponding decoded string.
    """
    @wraps(method)
    def method_wrapper(self, data: bytes) -> Iterable[bytes]:
        lines = data.decode(self.codec).splitlines()
        width = len(str(len(lines)))
        for k, line in enumerate(lines):
            try:
                yield method(self, line).encode(self.codec)
            except Exception as E:
                self.log_info(F'error in line {k:0{width}d}: {E}')
    return method_wrapper



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import json
import re

from .. import Unit
from ...lib.decorators import unicoded
from ...lib.argformats import number


class ppjson(Unit):
    """
    Expects JSON input data and outputs it in a neatly formatted manner.
    """
    _TRAILING_COMMA = R',\s*(}|])'

    def interface(self, argp):
        argp.add_argument('-i', '--indent', type=number, default=4,
            help='Controls the amount of space characters used for indentation in the output. Default is 4.')
        return super().interface(argp)

    @unicoded
    def process(self, data: str) -> str:
        if re.search(self._TRAILING_COMMA, data):
            from ...lib.patterns import formats

            strings = {
                range(*m.span())
                for m in re.finditer(formats.string.pattern, data)
            }

            def smartfix(match):
                k = match.start()
                return match.group(0 if any(k in s for s in strings) else 1)

            data = re.sub(self._TRAILING_COMMA, smartfix, data)

        return json.dumps(json.loads(data), indent=self.args.indent)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from Crypto.Cipher import Salsa20

from . import LatinStreamCipher


class salsa(LatinStreamCipher, cipher=Salsa20):
    """
    Salsa20 encryption and decryption. The default IV (nonce) value is the
    string value `REFINERY`.
    """
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from . import ArithmeticUnit


class rotl(ArithmeticUnit):
    """
    Rotate the bits of each block left.
    """
    def operate(self, value, shift):
        shift %= self.fbits
        return (value << shift) | (value >> (self.fbits - shift))

    def inplace(self, value, shift):
        shift %= self.fbits
        lower = value >> (self.fbits - shift)
        value <<= shift
        value |= lower



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

from ....lib.patterns import formats
from . import string_unquote, string_quote
from .. import Deobfuscator


class deob_ps1_format(Deobfuscator):
    """
    PowerShell deobfuscation for the following "format string"-based technique:

    - `"{0}{2}{1}"-f 'signa','ures','t'`
    - `"{0}na{2}{1}"-f 'sig','ures','t'`
    """
    def deobfuscate(self, data):
        pattern = R'\s*'.join([
            R'''(?P<pattern>{s})''',
            R'''-[fF]''',
            R'''(?P<args>({s},\s*)*{s})'''
        ]).format(s=formats.ps1str)
        pattern = re.compile(pattern)

        def deobfuscate(match):
            md = match.groupdict()
            self.log_debug(F'found match at {match.start()}: {match.group(0)[:30]}')
            pattern = string_unquote(md['pattern'])
            args = re.split(F'({formats.ps1str})', md['args'])
            args = [
                string_unquote(a.strip())
                for a in args[1::2]
            ]

            def argreplace(m):
                try:
                    index = int(m.group(1))
                    return args[index]
                except IndexError:
                    self.log_debug(F'only found {len(args)} arguments and format sequence {index}, aborting.')
                    raise
            try:
                return string_quote(re.sub(R'\{(\d+)\}', argreplace, pattern))
            except IndexError:
                return match.group(0)

        return pattern.sub(deobfuscate, data)



# -*- coding: utf-8 -*-
"""
Parsing of the .NET header. The code is based on the description in [1].

References:
  [1]: https://www.ntcore.com/files/dotnetformat.htm
"""
import pefile

from .types import (
    RawBytes,
    Box,
    Byte,
    NullTerminatedString,
    ParserException,
    StreamReader,
    StringGUID,
    StringPrimitive,
    Struct,
    UInt16,
    UInt32,
    UInt64,
    UnicodeString,
    unpack,
    ParserEOF
)


class InvalidDotNetHeader(ValueError):
    def __init__(self, msg=None):
        ValueError.__init__(self, msg or '.NET parsing failed: corrupt header.')


class InvalidSignature(InvalidDotNetHeader):
    def __init__(self):
        ValueError.__init__(self, '.NET parsing failed: Invalid signature.')


class BitMask:
    def __init__(self, bitmask):
        self._bitmask = bitmask

    def __contains__(self, pos):
        return self[pos] == 1

    def __len__(self):
        return self._bitmask.bit_length()

    def __getitem__(self, pos):
        return (self._bitmask >> pos) & 1

    def __iter__(self):
        for k in range(len(self)):
            if k in self:
                yield k

    def __repr__(self):
        return '{:b}'.format(self._bitmask)


def bits_required(n):
    return 0 if not n else (n - 1).bit_length()


class MultiTableIndex(Struct):
    _name_to_id = {
        'Assembly'               : 0x20,
        'AssemblyOS'             : 0x22,
        'AssemblyProcessor'      : 0x21,
        'AssemblyRef'            : 0x23,
        'AssemblyRefOS'          : 0x25,
        'AssemblyRefProcessor'   : 0x24,
        'ClassLayout'            : 0x0F,
        'Constant'               : 0x0B,
        'CustomAttribute'        : 0x0C,
        'ENCLog'                 : 0x1E,
        'ENCMap'                 : 0x1F,
        'Event'                  : 0x14,
        'EventMap'               : 0x12,
        'EventPtr'               : 0x13,
        'ExportedType'           : 0x27,
        'Field'                  : 0x04,
        'FieldLayout'            : 0x10,
        'FieldMarshal'           : 0x0D,
        'FieldPtr'               : 0x03,
        'FieldRVA'               : 0x1D,
        'File'                   : 0x26,
        'GenericParam'           : 0x2A,
        'GenericParamConstraint' : 0x2C,
        'ImplMap'                : 0x1C,
        'InterfaceImpl'          : 0x09,
        'ManifestResource'       : 0x28,
        'MemberRef'              : 0x0A,
        'MethodDef'              : 0x06,
        'MethodImpl'             : 0x19,
        'MethodPtr'              : 0x05,
        'MethodSemantics'        : 0x18,
        'MethodSpec'             : 0x2B,
        'Module'                 : 0x00,
        'ModuleRef'              : 0x1A,
        'NestedClass'            : 0x29,
        'Param'                  : 0x08,
        'ParamPtr'               : 0x07,
        'Permission'             : 0x0E,
        'Property'               : 0x17,
        'PropertyMap'            : 0x15,
        'PropertyPtr'            : 0x16,
        'StandAloneSig'          : 0x11,
        'TypeDef'                : 0x02,
        'TypeRef'                : 0x01,
        'TypeSpec'               : 0x1B,
    }
    refs = ()

    def __init__(self, reader, rows):
        if not self.refs:
            raise NotImplementedError
        row_max_len = max(
            rows.get(self._name_to_id[n], 0)
            for n in self.refs if n is not None
        )
        bits = self.bits + bits_required(row_max_len)
        Type = UInt32 if bits > 16 else UInt16
        Struct.__init__(self, reader, _type=Type)

    def __len__(self):
        return len(self.refs)

    @property
    def bits(self):
        return bits_required(len(self))

    @property
    def mask(self):
        return (1 << self.bits) - 1

    def parse(self):
        raw = self.expect(self._type)
        try:
            self.RowName = self.refs[raw & self.mask]
        except IndexError:
            self.Error = 'no ref at index {}'.format(raw & self.mask)
            self.RowName = None
        if self.RowName is not None:
            self.RowType = self._name_to_id[self.RowName]
        self.Index = raw >> self.bits


class TypeDefOrRefIndex(MultiTableIndex):
    refs = (
        'TypeDef',
        'TypeRef',
        'TypeSpec'
    )


class HasConstantIndex(MultiTableIndex):
    refs = (
        'Field',
        'Param',
        'Property'
    )


class HasCustomAttributeIndex(MultiTableIndex):
    refs = (
        'MethodDef',
        'Field',
        'TypeRef',
        'TypeDef',
        'Param',
        'InterfaceImpl',
        'MemberRef',
        'Module',
        'Permission',
        'Property',
        'Event',
        'StandAloneSig',
        'ModuleRef',
        'TypeSpec',
        'Assembly',
        'AssemblyRef',
        'File',
        'ExportedType',
        'ManifestResource',
    )


class HasFieldMarshallIndex(MultiTableIndex):
    refs = (
        'Field',
        'Param',
    )


class HasDeclSecurityIndex(MultiTableIndex):
    refs = (
        'TypeDef',
        'MethodDef',
        'Assembly',
    )


class MemberRefParentIndex(MultiTableIndex):
    refs = (
        'TypeDef',
        'TypeRef',
        'ModuleRef',
        'MethodDef',
        'TypeSpec',
    )


class HasSemanticsIndex(MultiTableIndex):
    refs = (
        'Event',
        'Property',
    )


class MethodDefOrRefIndex(MultiTableIndex):
    refs = (
        'MethodDef',
        'MemberRef',
    )


class MemberForwardedIndex(MultiTableIndex):
    refs = (
        'Field',
        'MethodDef',
    )


class ImplementationIndex(MultiTableIndex):
    refs = (
        'File',
        'AssemblyRef',
        'ExportedType',
    )


class CustomAttributeTypeIndex(MultiTableIndex):
    refs = (
        None,
        None,
        'MethodDef',
        'MemberRef',
        None
    )


class ResolutionScopeIndex(MultiTableIndex):
    refs = (
        'Module',
        'ModuleRef',
        'AssemblyRef',
        'TypeRef'
    )


class TypeOrMethodDefIndex(MultiTableIndex):
    refs = (
        'TypeDef',
        'MethodDef',
    )


class FieldIndex(MultiTableIndex):
    refs = ('Field',)


class MethodDefIndex(MultiTableIndex):
    refs = ('MethodDef',)


class ParamIndex(MultiTableIndex):
    refs = ('Param',)


class TypeDefIndex(MultiTableIndex):
    refs = ('TypeDef',)


class EventIndex(MultiTableIndex):
    refs = ('Event',)


class PropertyIndex(MultiTableIndex):
    refs = ('Property',)


class ModuleRefIndex(MultiTableIndex):
    refs = ('ModuleRef',)


class AssemblyRefIndex(MultiTableIndex):
    refs = ('AssemblyRef',)


class GenericParamIndex(MultiTableIndex):
    refs = ('GenericParam',)


class TableRow(Struct):
    def __init__(self, reader, streams, header, **kw):
        Struct.__init__(self, reader, _streams=streams, _header=header)

    def expect_strA(self):
        offset = self.expect(self._header.HeapOffsetTypes.String)
        return self._streams.Strings[offset]

    def expect_guid(self):
        index = self.expect(self._header.HeapOffsetTypes.GUID)
        offset = (index - 1) * 0x10
        return self._streams.GUID[offset]

    def expect_blob(self):
        offset = self.expect(self._header.HeapOffsetTypes.Blob)
        return self._streams.Blob[offset]

    def index(self, IndexParser):
        return self.expect(IndexParser, rows=self._header.RowCount)


class Module(TableRow):
    def parse(self):
        self.Generation = self.expect(UInt16)
        self.Name = self.expect_strA()
        self.MvId = self.expect_guid()
        self.EncId = self.expect_guid()
        self.EncBaseId = self.expect_guid()


class TypeRef(TableRow):
    def parse(self):
        self.ResolutionScope = self.index(ResolutionScopeIndex)
        self.TypeName = self.expect_strA()
        self.TypeNamespace = self.expect_strA()


class TypeDef(TableRow):
    def parse(self):
        self.Flags = self.expect(UInt32)
        self.TypeName = self.expect_strA()
        self.TypeNamespace = self.expect_strA()
        self.Extends = self.index(TypeDefOrRefIndex)
        self.FieldList = self.index(FieldIndex)
        self.MethodList = self.index(MethodDefIndex)


class FieldPtr(TableRow):
    def parse(self):
        self.Ref = self.expect(UInt16)


class Field(TableRow):
    def parse(self):
        self.Flags = self.expect(UInt16)
        self.Name = self.expect_strA()
        self.Signature = self.expect_blob()


class MethodPtr(TableRow):
    def parse(self):
        self.Ref = self.expect(UInt16)


class MethodDef(TableRow):
    def parse(self):
        self.RVA = self.expect(UInt32)
        self.ImplFlags = self.expect(UInt16)
        self.Flags = self.expect(UInt16)
        self.Name = self.expect_strA()
        self.Signature = self.expect_blob()
        self.ParamList = self.index(ParamIndex)


class ParamPtr(TableRow):
    def parse(self):
        self.Ref = self.expect(UInt16)


class Param(TableRow):
    def parse(self):
        self.Flags = self.expect(UInt16)
        self.Sequence = self.expect(UInt16)
        self.Name = self.expect_strA()


class InterfaceImpl(TableRow):
    def parse(self):
        self.Class = self.index(TypeDefIndex)
        self.Interface = self.index(TypeDefOrRefIndex)


class MemberRef(TableRow):
    def parse(self):
        self.Class = self.index(MemberRefParentIndex)
        self.Name = self.expect_strA()
        self.Signature = self.expect_blob()


class Constant(TableRow):
    def parse(self):
        self.Type = self.expect(UInt16)
        self.Parent = self.index(HasConstantIndex)
        self.Value = self.expect_blob()


class CustomAttribute(TableRow):
    def parse(self):
        self.Parent = self.index(HasCustomAttributeIndex)
        self.Type = self.index(CustomAttributeTypeIndex)
        self.Value = self.expect_blob()


class FieldMarshal(TableRow):
    def parse(self):
        self.Parent = self.index(HasFieldMarshallIndex)
        self.NativeType = self.expect_blob()


class Permission(TableRow):
    def parse(self):
        self.Action = self.expect(UInt16)
        self.Parent = self.index(HasDeclSecurityIndex)
        self.PermissionSet = self.expect_blob()


class ClassLayout(TableRow):
    def parse(self):
        self.PackingSize = self.expect(UInt16)
        self.ClassSize = self.expect(UInt32)
        self.Parent = self.index(TypeDefIndex)


class FieldLayout(TableRow):
    def parse(self):
        self.Offset = self.expect(UInt32)
        self.Field = self.index(FieldIndex)


class StandAloneSig(TableRow):
    def parse(self):
        self.Signature = self.expect_blob()


class EventMap(TableRow):
    def parse(self):
        self.Parent = self.index(TypeDefIndex)
        self.EventList = self.index(EventIndex)


class EventPtr(TableRow):
    def parse(self):
        self.Ref = self.expect(UInt16)


class Event(TableRow):
    def parse(self):
        self.EventFlags = self.expect(UInt16)
        self.Name = self.expect_strA()
        self.EventType = self.index(TypeDefOrRefIndex)


class PropertyMap(TableRow):
    def parse(self):
        self.Parent = self.index(TypeDefIndex)
        self.PropertyList = self.index(PropertyIndex)


class PropertyPtr(TableRow):
    def parse(self):
        self.Ref = self.expect(UInt16)


class Property(TableRow):
    def parse(self):
        self.Flags = self.expect(UInt16)
        self.Name = self.expect_strA()
        self.Type = self.expect_blob()


class MethodSemantics(TableRow):
    def parse(self):
        self.Semantics = self.expect(UInt16)
        self.Method = self.index(MethodDefIndex)
        self.Association = self.index(HasSemanticsIndex)


class MethodImpl(TableRow):
    def parse(self):
        self.Class = self.index(TypeDefIndex)
        self.MethodBody = self.index(MethodDefOrRefIndex)
        self.MethodDeclaration = self.index(MethodDefOrRefIndex)


class ModuleRef(TableRow):
    def parse(self):
        self.Name = self.expect_strA()


class TypeSpec(TableRow):
    def parse(self):
        self.Signature = self.expect_blob()


class ImplMap(TableRow):
    def parse(self):
        self.MappingFlags = self.expect(UInt16)
        self.MemberForwarded = self.index(MemberForwardedIndex)
        self.ImportName = self.expect_strA()
        self.ImportScope = self.index(ModuleRefIndex)


class FieldRVA(TableRow):
    def parse(self):
        self.RVA = self.expect(UInt32)
        self.Field = self.index(FieldIndex)


class Assembly(TableRow):
    def parse(self):
        self.HashAlgId = self.expect(UInt32)
        self.MajorVersion = self.expect(UInt16)
        self.MinorVersion = self.expect(UInt16)
        self.BuildNumber = self.expect(UInt16)
        self.RevisionNumber = self.expect(UInt16)
        self.Flags = self.expect(UInt32)
        self.PublicKey = self.expect_blob()
        self.Name = self.expect_strA()
        self.Culture = self.expect_strA()


class AssemblyProcessor(TableRow):
    def parse(self):
        self.Processor = self.expect(UInt32)


class AssemblyOS(TableRow):
    def parse(self):
        self.OsPlatformId = self.expect(UInt32)
        self.OsMajorVersion = self.expect(UInt32)
        self.OsMinorVersion = self.expect(UInt32)


class AssemblyRef(TableRow):
    def parse(self):
        self.MajorVersion = self.expect(UInt16)
        self.MinorVersion = self.expect(UInt16)
        self.BuildNumber = self.expect(UInt16)
        self.RevisionNumber = self.expect(UInt16)
        self.Flags = self.expect(UInt32)
        self.PublicKeyOrToken = self.expect_blob()
        self.Name = self.expect_strA()
        self.Culture = self.expect_strA()
        self.HashValue = self.expect_blob()


class AssemblyRefProcessor(TableRow):
    def parse(self):
        self.Processor = self.expect(UInt32)
        self.AssemblyRef = self.index(AssemblyRefIndex)


class AssemblyRefOS(TableRow):
    def parse(self):
        self.OsPlatformId = self.expect(UInt32)
        self.OsMajorVersion = self.expect(UInt32)
        self.OsMinorVersion = self.expect(UInt32)
        self.AssemblyRef = self.index(AssemblyRefIndex)


class File(TableRow):
    def parse(self):
        self.Flags = self.expect(UInt32)
        self.Name = self.expect_strA()
        self.HashValue = self.expect_blob()


class ExportedType(TableRow):
    def parse(self):
        self.Flags = self.expect(UInt32)
        self.TypeDefId = self.expect(UInt32)
        self.TypeName = self.expect_strA()
        self.TypeNamespace = self.expect_strA()
        self.Implementation = self.index(ImplementationIndex)


class ManifestResource(TableRow):
    def parse(self):
        self.Offset = self.expect(UInt32)
        self.Flags = self.expect(UInt32)
        self.Name = self.expect_strA()
        self.Implementation = self.index(ImplementationIndex)


class NestedClass(TableRow):
    def parse(self):
        self.NestedClass = self.index(TypeDefIndex)
        self.EnclosingClass = self.index(TypeDefIndex)


class GenericParam(TableRow):
    def parse(self):
        self.Number = self.expect(UInt16)
        self.Flags = self.expect(UInt16)
        self.Owner = self.index(TypeOrMethodDefIndex)
        self.Name = self.expect_strA()


class MethodSpec(TableRow):
    def parse(self):
        self.Method = self.index(MethodDefOrRefIndex)
        self.Instantiation = self.expect_blob()


class GenericParamConstraint(TableRow):
    def parse(self):
        self.Owner = self.index(GenericParamIndex)
        self.Constraint = self.index(TypeDefOrRefIndex)


class ENCLog(TableRow):
    def parse(self):
        self.Token = self.expect(UInt32)
        self.FuncCode = self.expect(UInt32)


class ENCMap(TableRow):
    def parse(self):
        self.Token = self.expect(UInt32)


class ImageDataDirectory(Struct):
    def parse(self):
        self.VirtualAddress = self.expect(UInt32)
        self.Size = self.expect(UInt32)


class NetDirectory(Struct):
    def parse(self):
        self.Size = self.expect(UInt32)
        self.MajorRuntimeVersion = self.expect(UInt16)
        self.MinorRuntimeVersion = self.expect(UInt16)
        self.MetaData = self.expect(ImageDataDirectory)
        self.Flags = self.expect(UInt32)
        self.EntryPointToken = self.expect(UInt32)
        self.Resources = self.expect(ImageDataDirectory)
        self.StringNameSignature = self.expect(ImageDataDirectory)
        self.CodeManagerTable = self.expect(ImageDataDirectory)
        self.VTableFixups = self.expect(ImageDataDirectory)
        self.ExportAddressTableJumps = self.expect(ImageDataDirectory)
        self.ManagedNativeHeader = self.expect(ImageDataDirectory)
        # Known Flags
        self.KnownFlags = dict(
            ILONLY=((self.Flags >> 0) & 1 == 1),
            REQUIRE_32BIT=((self.Flags >> 1) & 1 == 1),
            IL_LIBRARY=((self.Flags >> 2) & 1 == 1),
            STRONGNAMESIGNED=((self.Flags >> 3) & 1 == 1),
            NATIVE_ENTRYPOINT=((self.Flags >> 4) & 1 == 1),
            TRACKDEBUGDATA=((self.Flags >> 16) & 1 == 1)
        )


class NetMetaDataStreamEntry(Struct):
    def parse(self):
        self.VirtualAddress = self.expect(UInt32)
        self.Size = self.expect(UInt32)
        self.Name = self.expect(NullTerminatedString, align=4)


class NetMetaDataTablesHeader(Struct):
    def parse(self):
        Types = (UInt16, UInt32)
        self._Reserved1 = self.expect(UInt32)
        self.MajorVersion = self.expect(Byte)
        self.MinorVersion = self.expect(Byte)
        self.Flags = BitMask(self.expect(Byte))
        self.KnownFlags = dict(
            PADDING=bool(self.Flags[3]),
            DELTA_ONLY=bool(self.Flags[5]),
            LARGE_STRA=bool(self.Flags[0]),  # Strings require 4 byte offsets
            LARGE_GUID=bool(self.Flags[1]),  # GUIDs require 4 byte offsets
            LARGE_BLOB=bool(self.Flags[2]),  # Blobs require 4 byte offsets
            EXTRA_DATA=bool(self.Flags[6]),  # Extra data follows the row counts
            HAS_DELETE=bool(self.Flags[7])   # Certain tables can contain deleted rows.
        )
        self.HeapOffsetTypes = Box(
            String=Types[self.Flags[0]],
            GUID=Types[self.Flags[1]],
            Blob=Types[self.Flags[2]])
        self._Reserved2 = self.expect(Byte)
        self.ExistingRows = BitMask(self.expect(UInt64))
        self.SortedRows = BitMask(self.expect(UInt64))
        self.RowCount = {k: self.expect(UInt32) for k in self.ExistingRows}


class NetMetaDataTables(Struct):
    lookup = {
        0x00: Module,
        0x01: TypeRef,
        0x02: TypeDef,
        0x03: FieldPtr,
        0x04: Field,
        0x05: MethodPtr,
        0x06: MethodDef,
        0x07: ParamPtr,
        0x08: Param,
        0x09: InterfaceImpl,
        0x0A: MemberRef,
        0x0B: Constant,
        0x0C: CustomAttribute,
        0x0D: FieldMarshal,
        0x0E: Permission,
        0x0F: ClassLayout,
        0x10: FieldLayout,
        0x11: StandAloneSig,
        0x12: EventMap,
        0x13: EventPtr,
        0x14: Event,
        0x15: PropertyMap,
        0x16: PropertyPtr,
        0x17: Property,
        0x18: MethodSemantics,
        0x19: MethodImpl,
        0x1A: ModuleRef,
        0x1B: TypeSpec,
        0x1C: ImplMap,
        0x1D: FieldRVA,
        0x1E: ENCLog,
        0x1F: ENCMap,
        0x20: Assembly,
        0x21: AssemblyProcessor,
        0x22: AssemblyOS,
        0x23: AssemblyRef,
        0x24: AssemblyRefProcessor,
        0x25: AssemblyRefOS,
        0x26: File,
        0x27: ExportedType,
        0x28: ManifestResource,
        0x29: NestedClass,
        0x2A: GenericParam,
        0x2B: MethodSpec,
        0x2C: GenericParamConstraint,
    }

    def __init__(self, reader, streams):
        Struct.__init__(self, reader, _streams=streams)

    def parse(self):
        self.Header = self.expect(NetMetaDataTablesHeader)
        if self.Header.Flags[6]:
            self.ExtraData = self.expect(UInt32)
        for k, Type in self.lookup.items():
            setattr(self, repr(Type), [])
        for k in sorted(self.Header.RowCount):
            count = self.Header.RowCount[k]
            try:
                Type = self.lookup[k]
            except KeyError:
                raise RuntimeError('Cannot parse unknown table index 0x{:08X}; unable to continue parsing.')
            TypeEntries = getattr(self, repr(Type))
            for _ in range(count):
                Entry = self.expect(Type, streams=self._streams, header=self.Header)
                TypeEntries.append(Entry)

    def __getitem__(self, k):
        try:
            Type = self.lookup[k]
        except KeyError:
            return super(NetMetaDataTables, self).__getitem__(k)
        else:
            return getattr(self, repr(Type))


class ODict(dict):
    def in_sequence(self, k, default=None):
        for j, index in enumerate(sorted(self)):
            if j == k:
                return self[index]
        else:
            return default


class NetMetaDataStreamDummy(dict):
    def __init__(self, default=None):
        self._default = default
        dict.__init__(self)

    def __getitem__(self, offset):
        return self._default


class NetMetaDataStream(dict):
    def __init__(self, reader, type, default=None):
        dict.__init__(self)
        self._default = default
        self._reader = reader
        self._type = type
        self._cached = False
        self.read()

    def _next(self):
        return self._reader.expect(self._type)

    def __getitem__(self, offset):
        if offset < 0:
            return self._default
        try:
            return dict.__getitem__(self, offset)
        except KeyError:
            try:
                self._reader.seek(offset)
                item = self._next()
            except ParserException:
                pass
            else:
                self[offset] = item
                return item
        try:
            closest = max(key for key in self if key < offset)
        except ValueError:
            return None
        container = unpack(self[closest])
        return container[offset - closest:] or self._default

    def read(self):
        if self._cached:
            return
        self._reader.seek(0)
        while True:
            offset = self._reader.tell()
            try:
                self[offset] = self._next()
            except ParserException:
                break
        self._cached = True

    def __iter__(self):
        self.read()
        return dict.__iter__(self)


class NetMetaDataStreams(Struct):
    def __init__(self, reader, meta):
        Struct.__init__(self, reader, _meta=meta)

    def _read_all(self, reader, Type):
        while True:
            offset = reader.tell()
            try:
                yield offset, reader.expect(Type)
            except ParserException:
                break

    def parse(self):
        self.Tables = None
        self.Strings = NetMetaDataStreamDummy('')
        self.US = NetMetaDataStreamDummy('')
        self.GUID = NetMetaDataStreamDummy()
        self.Blob = NetMetaDataStreamDummy(B'')
        with self._reader.checkpoint():
            TableName = '#~'
            for se in self._meta.StreamInfo:
                if se.Name == TableName:
                    break
                if se.Name == '#-':
                    TableName = se.Name
                    break
            for k, name in reversed(tuple(enumerate((TableName, '#Strings', '#US', '#GUID', '#Blob')))):
                for _, Entry in enumerate(self._meta.StreamInfo):
                    if Entry.Name.upper() == name.upper():
                        break
                else:
                    continue
                self._reader.seek(Entry.VirtualAddress)
                try:
                    reader = StreamReader(self._reader.read(Entry.Size))
                except ParserEOF:
                    continue
                if name != TableName:
                    Default = ['', '', None, B''][k - 1]
                    Type = [NullTerminatedString, UnicodeString, StringGUID, RawBytes][k - 1]
                    Stream = NetMetaDataStream(reader, Type, Default)
                    setattr(self, name[1:], Stream)
                else:
                    self.Tables = reader.expect(NetMetaDataTables, streams=self)


class NetMetaData(Struct):
    @property
    def resources(self):
        return self.Streams.Tables.ManifestResource

    @property
    def RVAs(self):
        return self.Streams.Tables.FieldRVA

    def parse(self):
        try:
            self.Signature = self.expect(UInt32)
        except ParserEOF:
            raise InvalidSignature
        if self.Signature != 0x424A5342:
            raise InvalidSignature
        self.MajorVersion = self.expect(UInt16)
        self.MinorVersion = self.expect(UInt16)
        self._Reserved = self.expect(UInt32)
        size = self.expect(UInt32)
        self.VersionString = self.expect(StringPrimitive, size=size, align=4)
        self.Flags = self.expect(UInt16)
        self.StreamCount = self.expect(UInt16)
        self.StreamInfo = [
            self.expect(NetMetaDataStreamEntry)
            for _ in range(self.StreamCount)
        ]
        self.Streams = self.expect(NetMetaDataStreams, meta=self)


class NetResourceWithName(Struct):
    def parse(self):
        self.Name = self.expect(StringPrimitive, codec='utf-16LE')
        self.Offset = self.expect(UInt32)
        with self._reader.checkpoint():
            self._reader.seek(self.Offset)
            self.Size = self.expect(UInt32)
            self.Data = self._reader.read(self.Size)


class DotNetHeader:
    def __init__(self, data, parse_resources=True):
        try:
            self.pe = pefile.PE(data=data, fast_load=True)
            self.head = NetDirectory(self.reader(
                self.pe.OPTIONAL_HEADER.DATA_DIRECTORY[14]))
        except IndexError:
            # this is a temporary fix for what should really be handled in pefile,
            # see also: https://github.com/erocarrera/pefile/issues/264
            import struct
            count = self.pe.OPTIONAL_HEADER.NumberOfRvaAndSizes
            delta = (self.pe.FILE_HEADER.SizeOfOptionalHeader
                - (self.pe.OPTIONAL_HEADER.sizeof() + count * 8)) // 8
            if delta > 0:
                nt = self.pe.DOS_HEADER.e_lfanew
                data = data[:nt + 0x74] + struct.pack('<I', count + delta) + data[nt + 0x78:]
            self.pe = pefile.PE(data=data, fast_load=True)
            self.head = NetDirectory(self.reader(
                self.pe.OPTIONAL_HEADER.DATA_DIRECTORY[14]))
        except pefile.PEFormatError:
            raise InvalidDotNetHeader

        try:
            self.meta = NetMetaData(self.reader(self.head.MetaData))
        except pefile.PEFormatError:
            raise InvalidDotNetHeader

        if not parse_resources:
            self.resources = []
        else:
            self.parse_resources()

    def parse_resources(self):
        def parse(reader):
            for entry in self.meta.resources:
                try:
                    reader.seek(entry.Offset)
                    size = reader.expect(UInt32)
                    yield Box(
                        Name=entry.Name,
                        Data=reader.read(size)
                    )
                except ParserEOF:
                    yield Box(
                        Name=entry.Name,
                        Data=B''
                    )
        self.resources = list(parse(self.reader(self.head.Resources)))

    def reader(self, obj):
        return StreamReader(self.pe.get_data(obj.VirtualAddress, obj.Size))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from . import ArithmeticUnit


class xor(ArithmeticUnit):
    """
    Form the exclusive or of the input data with the given argument.
    """
    def process_ecb_fast(self, data):
        try:
            from Crypto.Util.strxor import strxor
        except ModuleNotFoundError:
            self.log_warn('the pycryptodome package does not seem to be installed, falling back to numpy.')
            return super().process_ecb_fast(data)
        from itertools import islice, cycle
        take = len(data) // self.args.blocksize + 1
        argb = self.unchunk(islice(cycle(x & self.fmask for x in self.args.arg[0]), take))
        return strxor(data, argb[:len(data)])

    @staticmethod
    def operate(a, b): return a ^ b
    @staticmethod
    def inplace(a, b): a ^= b



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re
from struct import pack
from string import printable

from .. import Unit


class esc(Unit):
    """
    Encodes and decodes common ASCII escape sequences.
    """
    _ESCAPE = {
        0x00: BR'\0',
        0x07: BR'\a',
        0x08: BR'\b',
        0x0C: BR'\f',
        0x0A: BR'\n',
        0x0D: BR'\r',
        0x09: BR'\t',
        0x0B: BR'\v',
        0x5C: BR'\\',
        0x27: BR'\'',
        0x22: BR'\"'
    }
    _UNESCAPE = {
        BR'0': B'\x00',
        BR'a': B'\x07',
        BR'b': B'\x08',
        BR'f': B'\x0C',
        BR'n': B'\x0A',
        BR'r': B'\x0D',
        BR't': B'\x09',
        BR'v': B'\x0B',
        B'\\': B'\x5C',
        BR"'": B'\x27',
        BR'"': B'\x22'
    }

    def interface(self, argp):
        mode = argp.add_mutually_exclusive_group()
        mode.add_argument('-x', '--hex', action='store_true',
            help='Hex encode everything, do not use C escape sequences.')
        mode.add_argument('-u', '--unicode', action='store_true',
            help='Expect input/output to be an UTF-8 encoded unicode string and '
                 'use unicode escape sequences.')
        return super().interface(argp)

    def process(self, data):
        if self.args.unicode:
            return data.decode('UNICODE_ESCAPE').encode(self.codec)

        def unescape(match):
            c = match.group(1)
            if c[0] == 0x75:  # unicode
                return pack('H', int(c[1:], 16))
            if c[0] == 0x78:  # hexadecimal
                return pack('B', int(c[1:], 16))
            return self._UNESCAPE.get(c, c)
        data = re.sub(
            RB'\\(u[a-fA-F0-9]{4}|x[a-fA-F0-9]{2}|.)', unescape, data)
        return data

    def reverse(self, data):
        if self.args.unicode:
            return data.decode(self.codec).encode('UNICODE_ESCAPE')

        def escape(c):
            if chr(c) not in printable or c in self._ESCAPE and self.args.hex:
                return RB'\x%02x' % c
            else:
                return self._ESCAPE.get(c, B'%c' % c)
        return B''.join(escape(c) for c in data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from ...lib.argformats import PythonExpression
from . import ArithmeticUnit, NoNumpy


_BLOCK_OPERATION_PARSER = PythonExpression('B', 'A', 'S', 'N', 'V')


def blockop_expression(definition):
    """
    An argparse type which uses the `refinery.lib.argformats.PythonExpression` parser to
    parse the expressions that can be passed to `refinery.blockop`. Essentially, these
    are Python expressions which can contain variables `B`, `A`, `S`, and `V`.
    """
    def wrapper(B, S, N, *V):
        return wrapper.parsed(B=B, A=V[0], N=N, S=S, V=V) if V else wrapper.parsed(B=B, S=S)
    wrapper.parsed = _BLOCK_OPERATION_PARSER(definition)
    return wrapper


class blockop(ArithmeticUnit):
    """
    This unit allows you to specify a custom Python expression where the following variables are allowed.

    - the variable `A`: same as `V[0]`
    - the variable `B`: current block
    - the variable `N`: number of bytes in the input
    - the variable `S`: an optional seed value for an internal state
    - the variable `V`: the vector of arguments

    Each block of the input is replaced by the value of this expression. Additionally, it is possible to
    specify prologue and epilogue expressions which are used to update the state variable `S` before and
    after the update of each block, respectively.
    """

    def interface(self, argp):
        argp.add_argument('operation', type=blockop_expression, help='A Python expression defining the operation.')
        argp.prologue = argp.add_mutually_exclusive_group()
        argp.prologue.add_argument('-p', '--prologue', type=blockop_expression, metavar='E', default=None, help=(
            'Optional expression with which the state variable S is updated before a block is operated on.'))
        argp.epilogue = argp.add_mutually_exclusive_group()
        argp.epilogue.add_argument('-e', '--epilogue', type=blockop_expression, metavar='E', default=None, help=(
            'Optional expression with which the state variable S is updated after a block was operated on.'))
        argp.epilogue.add_argument('--inc', action='store_const', dest='epilogue', const=blockop_expression('S+1'),
            help='equivalent to --epilogue=S+1')
        argp.epilogue.add_argument('--dec', action='store_const', dest='epilogue', const=blockop_expression('S-1'),
            help='equivalent to --epilogue=S-1')
        argp.epilogue.add_argument('--cbc', action='store_const', dest='epilogue', const=blockop_expression('B'),
            help='equivalent to --epilogue=B')
        argp.add_argument('-s', '--seed', type=PythonExpression('N'), default=0, help=(
            'Optional seed value for the state variable S. The default is zero. This can be an expression '
            'involving the variable N.'
        ))
        return super().interface(argp)

    @property
    def ecb(self):
        return not self.args.epilogue and not self.args.prologue

    def process_ecb_fast(self, data):
        if not self.ecb:
            raise NoNumpy
        return super().process_ecb_fast(data)

    def process(self, data):
        self._total = len(data)
        self._state = self.args.seed
        if callable(self._state):
            self._state = self._state(N=self._total)
        return super().process(data)

    def operate(self, block, *args):
        if self.args.prologue:
            self._state = self.args.prologue(block, self._state, self._total, *args)
        block = self.args.operation(block, self._state, self._total, *args) & self.fmask
        if self.args.epilogue:
            self._state = self.args.epilogue(block, self._state, self._total, *args)
        return block



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from Crypto.Cipher import ARC4

from . import StandardCipherUnit

ARC4.key_size = range(1, 257)


class rc4(StandardCipherUnit, cipher=ARC4):
    """
    RC4 encryption and decryption.
    """
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit


class cp1252(Unit):
    """
    Encodes and decodes Windows CP 1252 (aka Latin1) encoded string data.
    """

    def process(self, data):
        return data.decode(self.codec).encode('cp1252')

    def reverse(self, data):
        return data.decode('cp1252').encode(self.codec)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import base64

from .. import Unit


class b64(Unit):
    """
    Base64 encoding and decoding.
    """

    def interface(self, argp):
        argp.add_argument(
            '-u', '--urlsafe',
            dest='altchars',
            action='store_const',
            default=None,
            const=B'-_',
            help='use urlsafe special chars -_'
        )
        return super().interface(argp)

    def reverse(self, data):
        return base64.b64encode(data, altchars=self.args.altchars)

    def process(self, data):
        if len(data) <= 1:
            raise ValueError('single byte can not be base64-decoded.')
        if data:
            data += B'=' * (-len(data) % 4)
            return base64.b64decode(data, altchars=self.args.altchars)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Implements various hashing algorithms.
"""
from ... import Unit


class HashUnit(Unit, abstract=True):

    _algorithm = NotImplemented

    def interface(self, argp):
        argp.add_argument('-t', '--text', action='store_true',
            help='Output a hexadecimal representation of the hash')
        return super().interface(argp)

    def process(self, data: bytes) -> bytes:
        digest = self._algorithm(data)
        return digest.hex().encode(self.codec) if self.args.text else digest



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A package containing several sub-packages for various data formats.
"""
import fnmatch
import re

from .. import Unit
from ...lib.argformats import number, virtualaddr


def pathspec(expression):
    """
    Normalizes a path which is separated by backward or forward slashes to be
    separated by forward slashes.
    """
    return '/'.join(re.split(R'[\\\/]', expression))


class EndOfStringNotFound(ValueError):
    def __init__(self):
        super().__init__('end of string could not be determined')


class PathExtractorUnit(Unit):

    def interface(self, argp):
        argp.add_argument('paths', metavar='path', nargs='*', default=['*'], type=pathspec,
            help=(
                'A path from which data is to be extracted. Each item is returned '
                ' as a separate output of this unit. Paths may contain wildcards. '
                'The default is a single asterix, which means that every item will '
                'be extracted.'
            )
        )
        return super().interface(argp)

    def _check_reachable(self, path: str) -> bool:
        for pattern in self.args.path:
            stops = [k for k, c in enumerate(pattern) if c in '/*?'] + [None]
            for stop in stops:
                if fnmatch.fnmatch(path, pattern[:stop]):
                    return True

    def _check_path(self, path: str) -> bool:
        return any(fnmatch.fnmatch(path, pattern) for pattern in self.args.paths)


class MemoryExtractorUnit(Unit):

    def interface(self, argp):
        limit = argp.add_mutually_exclusive_group()
        limit.add_argument('-t', '--take', type=number[1:], default=0,
            help='The number of bytes to read.')
        limit.add_argument('-e', '--end', type=virtualaddr, default=None,
            help='Read bytes until this offset, which has to be located after the starting offset.')
        limit.add_argument('-a', '--ascii', action='store_true',
            help='Read the memory at the given offset as an ASCII string.')
        limit.add_argument('-u', '--utf16', action='store_true',
            help='Read the memory at the given offset as an UTF16 string.')
        argp.add_argument('offset', type=virtualaddr,
            help='Specify virtual offset as either .section:OFFSET or just a virtual address in hex.')
        return super().interface(argp)

    def _read_from_memory(self, data, offset_oracle):

        start, end = offset_oracle(self.args.offset)

        if self.args.end:
            end = offset_oracle(self.args.end)
            if end < start:
                raise ValueError(
                    F'The end offset 0x{end:08X} lies {start-end} bytes '
                    F'before the start offset 0x{start:08X}.'
                )
        elif self.args.take:
            end = start + self.args.take
        elif self.args.ascii:
            end = data.find(B'\0', start)
            if end < 0:
                raise EndOfStringNotFound
        elif self.args.utf16:
            for end in range(start, len(data), 2):
                if not data[end] and not data[end + 1]:
                    break
            else:
                raise EndOfStringNotFound

        return data[start:end]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from pefile import PE, PEFormatError, DIRECTORY_ENTRY
from struct import unpack

from ... import Unit


class carve_pe(Unit):
    """
    Extracts anything from the input data that looks like a Portable
    Executable (PE) file.
    """

    def _pesize(self, pe: PE) -> int:
        overlay = pe.get_overlay_data_start_offset() or 0
        maxaddr = max(
            s.PointerToRawData + s.SizeOfRawData
            for s in pe.sections
        )
        maxdata = max(
            pe.get_offset_from_rva(d.VirtualAddress) + d.Size
            for d in pe.OPTIONAL_HEADER.DATA_DIRECTORY
        )
        # The certificate overlay is given as a file offset
        # rather than a virtual address.
        cert = pe.OPTIONAL_HEADER.DATA_DIRECTORY[
            DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_SECURITY']]
        certend = cert.VirtualAddress + cert.Size
        self.log_debug(F'overlay at 0x{overlay:08X}')
        self.log_debug(F'maxaddr at 0x{maxaddr:08X}')
        self.log_debug(F'maxdata at 0x{maxdata:08X}')
        self.log_debug(F'certend at 0x{certend:08X}')
        return max(overlay, maxaddr, maxdata, certend)

    def process(self, data):
        cursor = 0
        while True:
            p = data.find(B'MZ', cursor)
            if p < cursor: break
            cursor = p + 2
            k, = unpack('H', data[p + 0x3C:p + 0x3E])
            if data[p + k:p + k + 2] != B'PE':
                continue
            try:
                pe = PE(data=data[p:])
            except PEFormatError as err:
                self.log_debug('parsing of PE header at 0x{p:08X} failed:', err)
                continue
            pesize = self._pesize(pe)
            self.log_info(F'extracted PE file of size 0x{pesize:08X} from 0x{p:08X}')
            yield data[p:p + pesize]
            cursor = pe.OPTIONAL_HEADER.SizeOfHeaders + p



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import html as html_

from .. import Unit
from ...lib.decorators import unicoded


class html(Unit):
    """
    Encodes and decodes common ASCII escape sequences.
    """

    @unicoded
    def process(self, data: str) -> str:
        return html_.unescape(data)

    @unicoded
    def reverse(self, data: str) -> str:
        return html_.escape(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit
from ...lib.argformats import number

from os.path import isfile
from glob import glob


class fread(Unit):
    """
    Reads files from disk and outputs them individually. Has the ability to
    read large files in chunks.
    """

    def interface(self, argp):
        argp.add_argument(
            '-s', '--size',
            metavar='N',
            type=number,
            default=None,
            help=(
                'If specified, files will be read in chunks of size N and each '
                'chunk is emitted as one element in the output list.'
            )
        )
        argp.add_argument(
            'filenames',
            nargs='+',
            metavar='FILEMASK',
            help=(
                'A list of file masks (with wildcard patterns). Each matching '
                'file will be read from disk and emitted.'
            )
        )
        return super().interface(argp)

    def _read_chunks(self, fd):
        while True:
            buffer = fd.read(self.args.size)
            if not buffer:
                break
            yield buffer

    def process(self, data):
        for mask in self.args.filenames:
            for filename in glob(mask, recursive=True):
                if not isfile(filename):
                    continue
                try:
                    with open(filename, 'rb') as stream:
                        if not self.args.size:
                            self.log_info('reading:', filename)
                            yield dict(data=stream.read(), path=filename)
                        else:
                            yield from self._read_chunks(stream)
                except PermissionError:
                    self.log_warn('permission denied:', filename)
                except FileNotFoundError:
                    self.log_warn('file is missing:', filename)
                except Exception:
                    self.log_warn('unknown error while reading:', filename)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

from . import RegexUnit


class resplit(RegexUnit):
    """
    Splits the data at the given regular expression and returns the sequence of
    chunks between the separators. By default, the input is split along line breaks.
    """
    def interface(self, argp):
        return super().interface(argp, regex_default=RB'\r?\n')

    def process(self, data):
        yield from re.split(self.args.regex, data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from Crypto.Cipher import Blowfish

from . import StandardCipherUnit


class blowfish(StandardCipherUnit, cipher=Blowfish):
    """
    Blowfish encryption and decryption.
    """
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from Crypto.Protocol.KDF import HKDF as HKDF_

from . import KeyDerivation


class HKDF(KeyDerivation):
    """HKDF Key derivation"""

    _DEFAULT_SALT = None
    _DEFAULT_HASH = 'SHA512'

    def process(self, data):
        return HKDF_(data, self.args.size, self.salt, self.algorithm)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Contains all units that can work on blocks a fixed length. Note that block cipher
algorithms can be found in `refinery.units.crypto.cipher`.
"""
from itertools import cycle
from inspect import signature, Parameter

from .. import Unit
from ...lib.argformats import numbin, number
from ...lib import chunks


class NoNumpy(Exception):
    pass


class BlockTransformation(Unit, abstract=True):
    def interface(self, argp):
        block = argp.add_argument_group(
            'Block Options',
            'Controls how the input data is split into blocks.'
        )
        block.add_argument('-E', '--bigendian', dest='little_endian', action='store_false',
            help='Read chunks in big endian.')
        block.add_argument('-B', '--blocksize', metavar='N', type=number[1:], default=1,
            help='The size of each block in bytes, default is 1.')

        return super().interface(argp)

    @property
    def bytestream(self):
        return self.args.blocksize == 1

    @property
    def fbits(self):
        return 8 * self.args.blocksize

    @property
    def fmask(self):
        return (1 << self.fbits) - 1

    def rest(self, data):
        if self.bytestream:
            return B''
        end = self.args.blocksize * (len(data) // self.args.blocksize)
        return data[end:]

    def chunk(self, data, raw=False):
        if not raw:
            return chunks.unpack(data, self.args.blocksize, self.args.little_endian)

        def chunkraw(data):
            stop = len(data)
            stop = stop - stop % self.args.blocksize
            for k in range(0, stop, self.args.blocksize):
                yield data[k : k + self.args.blocksize]

        return chunkraw(data)

    def unchunk(self, data, raw=False):
        if not raw:
            return chunks.pack(data, self.args.blocksize, self.args.little_endian)

        def bytefilter(it):
            for item in it:
                if isinstance(item, (bytes, bytearray)):
                    yield item
                else:
                    yield bytes(item)

        return B''.join(bytefilter(data))

    def process(self, data):
        return self.unchunk(
            self.process_block(b) for b in self.chunk(data)
        ) + self.rest(data)

    def process_block(self, block):
        """
        A blockwise operation implements this routine to process each block, which
        is given as an integer. The return value is also expected to be an integer.
        """
        raise NotImplementedError


class ArithmeticUnit(BlockTransformation, abstract=True):
    operate = NotImplemented
    inplace = NotImplemented

    def interface(self, argp):
        specs = signature(self.operate)
        nargs = len(specs.parameters) - 1
        if any(p.kind == Parameter.VAR_POSITIONAL for p in specs.parameters.values()):
            nargs = '*'
        if nargs:
            argp.add_argument('arg', type=numbin, nargs=nargs, help=(
                'A single numeric expression which provides the right argument to the operation, '
                'where the left argument is each block in the input data. This argument can also '
                'contain a sequence of bytes which is then split into blocks of the same size as '
                'the input data and used cyclically.'))
        else:
            argp.set_defaults(arg=[])
        return super().interface(argp)

    def process_ecb_fast(self, data):
        """
        Attempts to perform the operation more quickly by using numpy arrays.
        """
        try:
            import numpy
        except ModuleNotFoundError:
            raise NoNumpy

        order = '<' if self.args.little_endian else '>'
        dtype = numpy.dtype(F'{order}u{self.args.blocksize}')
        blocks = len(data) // self.args.blocksize

        def nparg(buffer):
            if hasattr(buffer, '__len__') and len(buffer) == 1:
                buffer = buffer[0]
            if isinstance(buffer, int):
                self.log_info('detected numeric argument')
                return buffer & self.fmask
            return numpy.fromiter(cycle(buffer), dtype, blocks)

        rest = data[blocks * self.args.blocksize:]
        data = numpy.frombuffer(data, dtype, blocks)
        args = [nparg(a) for a in self.args.arg]

        if self.inplace is NotImplemented:
            data = self.operate(data, *args)
            if data.dtype != dtype:
                data = data.astype(dtype)
        else:
            data = data.copy()
            self.inplace(data, *args)
        return data.tobytes() + rest

    def process(self, data):
        try:
            result = self.process_ecb_fast(data)
        except NoNumpy:
            pass
        except Exception as error:
            self.log_warn(F'falling back to default method after numpy failed with error: {error}')
        else:
            self.log_debug('successfully used numpy to process data in ecb mode')
            return result
        self._arg = [cycle(a) for a in self.args.arg]
        return super().process(data)

    def process_block(self, block):
        return self.operate(block, *(next(a) & self.fmask for a in self._arg)) & self.fmask



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pefile

from .. import MemoryExtractorUnit


class peslice(MemoryExtractorUnit):
    """
    Extract data from PE files based on virtual offsets.
    """

    def _get_file_offset(self, pe, offset):
        addr = offset.address
        end = None
        if offset.section:
            name = offset.section.encode('latin-1')
            for section in pe.sections:
                if section.Name.find(name) in (0, 1, 2):
                    addr += section.PointerToRawData
                    end = addr + section.SizeOfRawData
                    self.log_debug('found section', name, F'at offset 0x{addr:08X}')
                    break
            else:
                raise ValueError(F'section {offset.section} was not found.')
        else:
            addr = pe.get_offset_from_rva(addr - pe.OPTIONAL_HEADER.ImageBase)
        return addr, end

    def process(self, data):
        try:
            pe = pefile.PE(data=data, fast_load=True)
        except Exception:
            raise ValueError('unable to parse input as PE file')

        return self._read_from_memory(data,
            lambda addr: self._get_file_offset(pe, addr))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from Crypto.Cipher import CAST

from . import StandardCipherUnit


class cast(StandardCipherUnit, cipher=CAST):
    """
    CAST encryption and decryption.
    """
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A simple tool to output binary data. Multiple arguments are output in framed
format, see `refinery.lib.frame`.
"""
from ...lib.argformats import multibin
from .. import Unit


class emit(Unit):

    def interface(self, argp):
        argp.add_argument('data', type=multibin, default=None, nargs='*',
            help=(
                'Data to be emitted. If no argument is specified, data '
                'is retrieved from the clipboard. Multiple arguments are '
                'output in framed format.'
            )
        )
        return super().interface(argp)

    def process(self, data):
        if not self.args.data:
            import pyperclip
            data = pyperclip.paste()
            try:
                data = data.encode(self.codec)
            except AttributeError:
                # data is None
                data = B''
            except UnicodeDecodeError:
                data = data.encode('utf-16le')
            yield data
        else:
            yield from self.args.data

    @classmethod
    def run(cls, argv=None, stream=None):
        super(emit, cls).run(argv=argv,
            stream=stream or open(__import__('os').devnull, 'rb')
        )



import refinery as r


class carve_b64(r.Unit):
    """
    Carves the longest base64 encoded expression and decodes it.
    """
    def process(self, data):
        return (r.carve('-lt1', 'b64') | r.b64)(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import io
import zipfile

from . import PathExtractorUnit


class xtzip(PathExtractorUnit):
    """
    Extract files from a Zip archive.
    """
    def process(self, data):
        with io.BytesIO(data) as stream:
            with zipfile.ZipFile(stream) as archive:
                for info in archive.infolist():
                    if info.is_dir():
                        continue
                    if self._check_path(info.filename):
                        self.log_info(info.filename)
                        yield dict(path=info.filename, data=archive.read(info.filename))
                    else:
                        self.log_debug(info.filename)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from Crypto.Cipher import DES

from . import StandardCipherUnit


class des(StandardCipherUnit, cipher=DES):
    """
    DES encryption and decryption.
    """
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re
import codecs

from email.header import decode_header

from .. import Unit
from ...lib.decorators import unicoded


class mimewords(Unit):
    """
    Implements the decoding of MIME encoded-word syntax from RFC-2047.
    """

    @unicoded
    def process(self, data: str) -> str:
        def replacer(match):
            self.log_info('encoded mime word:', match.group(0))
            decoded, = decode_header(match.group(0))
            raw, codec = decoded
            return codecs.decode(raw, codec, errors='surrogateescape')
        return re.sub(R"=(?:\?[^\?]*){3}\?=", replacer, data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

from .. import Deobfuscator


class deob_ps1_escape(Deobfuscator):
    def deobfuscate(self, data):
        return re.sub(R'''`([^0abfnrtv`#'"\$])''', R'\1', data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from string import whitespace
import re

from .. import Unit
from ...lib.argformats import multibin


class trim(Unit):
    """
    Removes byte sequences at beginning and end of input data.
    """

    def interface(self, argp):
        argp.add_argument('junk', type=multibin,
            default=[w.encode('ascii') for w in whitespace],
            help='Character(s) to be removed, default is whitespace.', nargs='*')
        one_side_only = argp.add_mutually_exclusive_group()
        one_side_only.add_argument('-l', '--left-only', action='store_true')
        one_side_only.add_argument('-r', '--right-only', action='store_true')
        return super().interface(argp)

    def process(self, data):
        keep_running = True
        while keep_running:
            keep_running = False
            for junk in self.args.junk:
                jlen = len(junk)
                if not self.args.right_only:
                    if data.startswith(junk):
                        if jlen == 1:
                            data = data.lstrip(junk)
                        else:
                            pattern = B'^(?:' + B''.join(B'\\x%02X' % X for X in junk) + B')+'
                            match = re.search(pattern, data)
                            data = data[match.end():]
                        keep_running = True
                if not self.args.left_only:
                    if data.endswith(junk):
                        if jlen == 1:
                            data = data.rstrip(junk)
                        else:
                            pattern = B'(?:' + B''.join(B'\\x%02X' % X for X in junk) + B')+$'
                            match = re.search(pattern, data)
                            data = data[:match.start()]
                        keep_running = True
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from . import FrameSlicer


class scope(FrameSlicer):
    """
    After using `refinery.scope` within in a `refinery.lib.frame`, all the
    following operations will be applied only to the selected indices. All
    remaining chunks still exist, they are just not operated on. When the
    frame closes or the frame is being rescoped by a second application of
    this unit, they become visible again.
    """
    def interface(self, argp):
        argp.add_argument('-n', '--not', dest='visible', action='store_false',
            help='Hide the given chunks instead of making them the only ones visible.')
        return super().interface(argp)

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        # Sort any slices with negative arguments to the back so we check
        # them last. This delays potential consumption of the chunks iterator
        # as much as possible.
        self.args.slice.sort(
            key=lambda s: (s.start or 0, s.stop or 0), reverse=True)

    def filter(self, chunks):
        it = iter(chunks)
        consumed = None
        size = None

        def buffered():
            yield from it
            while consumed:
                yield consumed.popleft()

        def shift(offset, default):
            nonlocal consumed, it, size
            if offset is None:
                return default
            if offset >= 0:
                return offset
            if consumed is None:
                from collections import deque
                self.log_info(F'consuming iterator to compute negative offset {offset}.')
                consumed = deque(it)
                size = len(consumed) + k + 1
            return max(0, offset + size)

        for k, chunk in enumerate(buffered()):
            for s in self.args.slice:
                if k in range(shift(s.start, 0), shift(s.stop, k + 1), s.step or 1):
                    chunk.visible = self.args.visible
                    break
            else:
                chunk.visible = not self.args.visible
            self.log_debug(chunk)
            yield chunk



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Exports two singletons `refinery.lib.types.INF` and `refinery.lib.types.AST`.
Used by `refinery.units.pattern.PatternExtractorBase` as the default values
for certain command line arguments.
"""
__all__ = ['INF', 'AST']


class INF:
    def __lt__(self, other): return False
    def __le__(self, other): return False
    def __gt__(self, other): return True
    def __ge__(self, other): return True
    def __eq__(self, other): return other is INF
    def __repr__(self): return '∞'


class AST:
    def __eq__(self, other): return True
    def __ne__(self, other): return False
    def __or__(self, other): return other
    def __repr__(self): return '*'


INF = INF()
"""
A crude object representing infinity, which is greater than anything it
is compared to, and only equal to itself.
"""

AST = AST()
"""
A wildcard object which is equal to everything.
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

from ... import Unit


class deob_vba_chr_literals(Unit):
    def process(self, data):
        def _chr(m):
            code = int(m.group(1), 0)
            if code == 34:
                return B'""""'
            return B'"%s"' % chr(code).encode('unicode_escape')
        data = re.sub(BR'Chr\((\d+x?\d+)\)', _chr, data, flags=re.IGNORECASE)
        data = re.sub(BR'"\s*\&\s*"', B'', data)
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from . import BlockTransformation
from ...lib.argformats import multibin


class map(BlockTransformation):
    """
    Each block of the input data which occurs as a block of the index argument
    is replaced by the corresponding block of the image argument.
    """
    def interface(self, argp):
        argp.add_argument('index', type=multibin, help='index characters')
        argp.add_argument('image', type=multibin, help='image characters')
        return super().interface(argp)

    def process(self, data):
        self._map = dict(zip(
            self.chunk(self.args.index),
            self.chunk(self.args.image)))
        return super().process(data)

    def process_block(self, token):
        return self._map.get(token, token)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import zlib

from .. import Unit
from ...lib.argformats import number


class zl(Unit):
    """
    ZLib compression and decompression.
    """

    def interface(self, argp):
        argp.add_argument('-l', '--level', action='store', type=number[0:9], default=9,
            help='specify level manually')
        argp.add_argument('-w', '--window', action='store', type=number[8:15], default=15,
            help='manually specify window size (but why.)')
        argp.add_argument('-f', '--force', action='store_true', help='decompress even if all known methods fail')
        mode = argp.add_mutually_exclusive_group()
        mode.add_argument('-z', '--zlib-header', action='store_true', help='use a zlib header')
        mode.add_argument('-g', '--gzip-header', action='store_true', help='use a gzip header')
        return super().interface(argp)

    def _force_decompress(self, data, mode):
        z = zlib.decompressobj(mode)

        def as_many_as_possible():
            for k in range(len(data)):
                try: yield z.decompress(data[k : k + 1])
                except zlib.error: break

        return B''.join(as_many_as_possible())

    def process(self, data):
        if data[0] == 0x78 or data[0:2] == B'\x1F\x8B' or self.args.zlib_header or self.args.gzip_header:
            mode_candidates = [self.args.window | 0x20, -self.args.window, 0]
        else:
            mode_candidates = [-self.args.window, self.args.window | 0x20, 0]
        for mode in mode_candidates:
            self.log_info(F'using mode {mode:+2d} for decompression')
            try:
                z = zlib.decompressobj(mode)
                return z.decompress(data)
            except zlib.error:
                pass
        if self.args.force:
            return self._force_decompress(data, mode_candidates[0])
        raise ValueError('could not detect any zlib stream.')

    def reverse(self, data):
        mode = -self.args.window
        if self.args.zlib_header:
            mode = -mode
        if self.args.gzip_header:
            mode = -mode | 0x10
        self.log_info(F'using mode {mode:+2d} for compression')
        zl = zlib.compressobj(self.args.level, zlib.DEFLATED, mode)
        zz = zl.compress(data)
        return zz + zl.flush(zlib.Z_FINISH)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from .. import Unit
from ...lib.argformats import number


class chop(Unit):
    """
    Reinterprets the input as a sequence of equally sized chunks and outputs
    this sequence.
    """

    def interface(self, argp):
        argp.add_argument('-t', '--truncate', action='store_true',
            help=(
                'Truncate possible excess bytes at the end of the input, '
                'by default they are appended as a single chunk.'
            )
        )
        argp.add_argument('-l', '--len', action='store_true',
            help=(
                'If this flag is specified, the size parameter determines '
                'the number of blocks to be produced rather than the size '
                'of each block.'
            )
        )
        argp.add_argument('size', type=number[1:],
            help='Chop data into chunks of this size.')

        return super().interface(argp)

    def process(self, data):
        size = self.args.size
        if self.args.len:
            size = len(data) // size
        for chunk in zip(*[iter(data)] * size):
            yield bytes(chunk)
        excess = len(data) % size
        if excess and not self.args.truncate:
            yield data[-excess:]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# Multibin Syntax

## Introduction

Many refinery units receive arguments which represent binary data, and usually these arguments can
be given in **multibin** format, which is a special syntax which allows to preprocess data with a
number of **handlers**. For example, the multibin expression `md5:password` preprocesses the
argument `password` (which is understood as its UTF8 encoding by default) using the `md5` handler,
which returns the MD5 hash of the input data. Consequently, the output of

    emit md5:password | hex -R

would be the string `5F4DCC3B5AA765D61D8327DEB882CF99`. The most important basic handlers to know
are:

- `s:string` disables all further preprocessing and interprets `string` as an UTF8 encoded string
- `u:string` same, but as an UTF16-LE encoded string
- `h:string` assumes that `string` is a hexadecimal string of even length and returns the decoded byte sequence.
- any unit's name can be prefixed to the string, i.e. `esc:\n` corresponds to the line break character.

If a multibin argument does not use any handler, refinery first interprets the string as the path
of an existing file on disk and attempts to return the contents of this file. If this fails, the
UTF8 encoding of the string is returned.

The handlers `copy` and `cut` as well as their shortcuts `c` and `x` are **final** handlers like
the above example `s`, i.e. the string that follows `copy:` will not be interpreted as a multibin
expression. Indeed, `copy` and `cut` expect the remaining string to be in Python slice syntax. The
expression `copy:0:1` would, for example, represent the first byte of the input data. With `copy`,
this data is copied out of the input and used for the argument. With `cut`, this data is removed
from the input data and used for the argument. All `cut` operations are performed in the order in
which the arguments are specified on the command line. For example, `emit Test | cca x::1 x::1`
will output `stTe`.

The modifiers `s`, `u`, `h`, `copy` (or `c`), and `cut` (or `x`) along with using unit modifiers
should cover most use cases. To learn about other existing modifiers, refer to the rest of this
documentation.

## The Details

This module implements all argument parser types for the binary refinery. Notable classes for the
command line use are the following:

- `refinery.lib.argformats.DelayedBinaryArgument` (used almost everywhere)
- `refinery.lib.argformats.DelayedNumbinArgument` (used by children of `refinery.units.blockwise.ArithmeticUnit`)
- `refinery.lib.argformats.DelayedRegexpArgument` (used by `refinery.rex`, `refinery.resub`)

All of the above classes inherit from `refinery.lib.argformats.DelayedArgument`. The following
mainly applies to `refinery.lib.argformats.DelayedBinaryArgument`, but the other two parsers work
similar. The classes implement the various modifiers which are available to multibin expressions.

The reason why these parsers have **"delayed"** in their name is that they allow the implementation
of handlers which require input data to be present, like the handlers `copy` and `cut`, which are
implemented in `refinery.lib.argformats.DelayedBinaryArgument.copy` and
`refinery.lib.argformats.DelayedBinaryArgument.cut`, respectively. These expressions can not be
evaluated immediately after the command line is parsed, but only as soon as input data becomes
available for processing.

In addition to the handlers which are implemented here, each refinery unit defines a (non-final)
modifier. For example, the expression `b64:Zm9v` corresponds to the binary string `foo`: The unit
`refinery.b64` is used to decode the string `Zm9v` here. Arguments can be passed to units in square
brackets and separated by commas, but there is no support for escaping comma characters. For
example, the multibin expression `xor[0xAA]:b64:2c/J2M/e` will return the binary string `secret` as
the final expression `2c/J2M/e` is base64-decoded and each byte xor'ed with the key `0xAA`. As a
second example, the expression

    hex[-R]:sha256:file:foobar.txt

will be parsed as the hexadecimal representation of the SHA256 hash of the file `foobar.txt`.
"""
import ast

from itertools import cycle, count, chain
from argparse import ArgumentTypeError
from contextlib import suppress
from functools import update_wrapper, wraps
from typing import Optional, Tuple, Union, Mapping, Any, List, Iterable, Callable


class PythonExpression:
    """
    Implements a parser for any Python expression with a prescribed set of variable
    names permitted to occur in the expression. The resulting object is a callable
    which can be given the string representation of such an expression. In turn, the
    result of this operation is either the value of the expression if no variables
    were present, or a callable which expects keyword arguments corresponding to the
    permitted variable names.
    """

    _ALLOWED_NODE_TYPES = {
        ast.Add,
        ast.BinOp,
        ast.BitAnd,
        ast.BitAnd,
        ast.BitOr,
        ast.BitXor,
        ast.BoolOp,
        ast.Compare,
        ast.Constant,
        ast.Div,
        ast.Eq,
        ast.FloorDiv,
        ast.Gt,
        ast.GtE,
        ast.IfExp,
        ast.Index,
        ast.Invert,
        ast.Is,
        ast.IsNot,
        ast.Load,
        ast.LShift,
        ast.Lt,
        ast.LtE,
        ast.List,
        ast.MatMult,
        ast.Mod,
        ast.Mult,
        ast.Name,
        ast.Not,
        ast.NotEq,
        ast.Num,
        ast.Or,
        ast.Pow,
        ast.RShift,
        ast.Slice,
        ast.Sub,
        ast.Subscript,
        ast.Tuple,
        ast.UAdd,
        ast.UnaryOp,
        ast.USub
    }

    def __init__(self, *variables):
        self.variables = set(variables)

    def __call__(self, definition):
        try:
            expression = ast.parse(definition)
            nodes = ast.walk(expression)
        except Exception:
            raise ArgumentTypeError('the provided expression could not be parsed')

        if type(next(nodes)) != ast.Module:
            raise ArgumentTypeError('unknown error parsing the expression')

        if type(next(nodes)) != ast.Expr:
            raise ArgumentTypeError('operation string is not a valid Python expression')

        nodes = list(nodes)
        types = set(type(node) for node in nodes)
        names = set(node.id for node in nodes if type(node) == ast.Name)

        if not types <= self._ALLOWED_NODE_TYPES:
            raise ArgumentTypeError(
                'the following operations are not allowed: {}'.format(
                    ', '.join(t.__name__ for t in types - self._ALLOWED_NODE_TYPES))
            )

        if not names <= self.variables:
            raise ArgumentTypeError(
                'the following variable names are unknown: {}'.format(
                    ', '.join(names - self.variables))
            )

        if not self.variables:
            return eval(definition)
        else:
            def evaluator(**kw): return eval(definition, None, kw)
            return evaluator


_PYTHON_EXPRESSION = PythonExpression()


def sliceobj(expression: Union[int, str]) -> slice:
    """
    Uses `refinery.lib.argformats.PythonExpression` to parse slice expressions
    where the bounds can be given as arithmetic expressions. For example, this
    argument format type will process the string `0x11:0x11+4*0x34` as the slice
    object `slice(17, 225, None)`.
    """
    if isinstance(expression, int):
        sliced = (expression,)
    else:
        sliced = expression.split(':')
        if not sliced or len(sliced) > 3:
            raise ArgumentTypeError(F'the expression {expression} is not a valid slice.')
        sliced = [None if not t else _PYTHON_EXPRESSION(t) for t in sliced]
    if len(sliced) == 1:
        k = sliced[0]
        return slice(k, k + 1) if k + 1 else slice(k, None, None)
    return slice(*sliced)


class virtualaddr:
    """
    Represents a virtual address; used by `refinery.peslice` and
    `refinery.elfslice` to reference offsets in executable file
    formats as they would appear in memory.
    """
    def __init__(self, address='0'):
        try:
            self.section, address = address.split(':')
        except ValueError:
            self.section = None
        else:
            if not address:
                self.address = 0
                return
        try:
            address = address.upper()
            if address.endswith('H'):
                address = '0x' + address[:-1]
            self.address = number[0:](address)
        except ValueError:
            pass
        try:
            self.address = int(address, 0x10)
        except ValueError:
            raise ArgumentTypeError(F'could not parse {address} as hexadecimal integer')


def utf8(x):
    """
    Returns the UTF8 encoding of the given string.
    """
    return x.encode('UTF8')


class number:
    __name__ = 'number'

    def __init__(self, min=None, max=None):
        self.min = min
        self.max = max

    def __getitem__(self, bounds):
        return self.__class__(bounds.start, bounds.stop)

    def __call__(self, value):
        if not isinstance(value, int):
            try:
                value = _PYTHON_EXPRESSION(value)
            except Exception:
                raise ValueError('unable to parse expression')
            if not isinstance(value, int):
                raise ArgumentTypeError('the expression with value {} is not an integer'.format(value))
        if self.min is not None and value < self.min or self.max is not None and value > self.max:
            raise ValueError('value {} is out of bounds [{}, {}]'.format(value, self.min, self.max))
        return value


number = number()
"""
The singleton instance of a class that uses `refinery.lib.argformats.PythonExpression`
to parse expressions with integer value. This singleton can be slice accessed to
create new number parsers, e.g. `number[0:]` will refuse to parse negative integer
expressions.
"""


class IncompatibleHandler(ValueError):
    """
    This exception is generated when `refinery.lib.argformats.DelayedArgument` handlers
    are chained in an incompatible way.
    """
    def __init__(self, type_expected, type_observed, modifier):
        self.type_expected = type_expected
        self.type_observed = type_observed
        self.modifier = modifier
        modifier_name = F'handler {modifier}' if modifier else 'default handler'
        super().__init__('{} received {} but expected {}'.format(
            modifier_name,
            type_observed.__name__,
            type_expected.__name__
        ))


class TooLazy(Exception):
    """
    Exception which indicates that an argument parser requires input data before it can be
    evaluated.
    """
    pass


class LazyEvaluation:
    """
    Empty parent class for any unit that throws `refinery.lib.argformats.TooLazy`.
    """
    pass


class DelayedArgumentDispatch:
    """
    This class is used as a decorator for the default handler of classes that inherit from
    `refinery.lib.argformats.DelayedArgument`. After decorating the routine `handler` with
    `refinery.lib.argformats.DelayedArgumentDispatch`, `handler.register` can be used to
    register additional handlers.
    """
    def __init__(self, method):
        update_wrapper(self, method)
        self.default = method
        self.handlers = {}
        self.final = {}
        self.units = {}

    def _get_unit(self, name, *args):
        name = name.replace('-', '_')
        uhash = hash((name,) + args)
        if uhash in self.units:
            return self.units[uhash]
        try:
            unit = getattr(__import__('refinery', None, None, [name]), name, None)
        except ModuleNotFoundError:
            return None
        else:
            unit = unit and unit(*args)
            self.units[uhash] = unit
            return unit

    def __get__(self, instance, instancetype):
        # We do not know the class whose methods we are decorating.
        self.instance = instance
        return self

    def __call__(self, data, modifier=None, *args):
        try:
            handler = self.default if modifier is None else self.handlers[modifier]
            return handler(self.instance, data, *args)
        except KeyError:
            import io
            unit = self._get_unit(modifier, *args)
            if not unit:
                raise ArgumentTypeError(F'failed to build unit {modifier}')
            with io.BytesIO(data) as stream:
                return B''.join(stream | unit)

    def can_handle(self, modifier, *args):
        return modifier in self.handlers or bool(self._get_unit(modifier, *args))

    def terminates(self, modifier):
        """
        Indicates whether the given registered modifier is final.
        """
        return self.final.get(modifier, False)

    def register(self, *modifiers, final=False):
        """
        Registers a new modifier handler.
        """
        def _register(method):
            for modifier in modifiers:
                self.handlers[modifier] = method
                self.final[modifier] = final
            return method
        return _register


class DelayedArgument(LazyEvaluation):
    """
    This base class for delayed argument parsers implements parsing
    expressions into supported modifiers.
    """
    _ARG_BEGIN_TOKEN = '['
    _ARG_CLOSE_TOKEN = ']'
    _ARG_SPLIT_TOKEN = ','

    def __init__(self, expression: str):
        self.modifiers = []
        self.finalized = False
        if not isinstance(self.handler, DelayedArgumentDispatch):
            raise NotImplementedError(
                'The default handler is required to be a '
                'DelayedArgumentDispatch instance.'
            )
        while not self.finalized:
            name, arguments, newexpr = self._split_modifier(expression)
            if not name or not self.handler.can_handle(name, *arguments):
                break
            self.modifiers.append((name, arguments))
            expression = newexpr
            if self.handler.terminates(name):
                self.finalized = True
        self.seed = expression
        self.modifiers.reverse()

    def _split_modifier(self, expression: str) -> Tuple[Optional[str], Tuple[str], str]:
        brackets = 0
        name = None
        argoffset = 0
        arguments = ()
        for k, character in enumerate(expression):
            if character == self._ARG_BEGIN_TOKEN:
                if not brackets:
                    if argoffset:
                        # This is the second time we encounter what appears to be an
                        # argument list, before the modifier has ended. This is not
                        # possible, and we decide to assume that no modifier was used.
                        break
                    name = expression[:k]
                    argoffset = k + 1
                brackets += 1
                continue
            if character == self._ARG_CLOSE_TOKEN:
                if brackets == 1:
                    arguments += expression[argoffset:k],
                elif not brackets:
                    if argoffset:
                        raise ArgumentTypeError(
                            F'Unable to parse {expression}, too many closing brackets.'
                        )
                    else:
                        break
                brackets -= 1
                continue
            if character == self._ARG_SPLIT_TOKEN:
                if brackets == 1:
                    arguments += expression[argoffset:k],
                    argoffset = k + 1
            if character == ':' and not brackets:
                if name is None:
                    name = expression[:k]
                return name, arguments, expression[k + 1:]
        return None, (), expression

    def __call__(self, data: Optional[bytearray] = None) -> bytes:
        arg = self.seed
        mod = iter(self.modifiers)
        if not self.finalized:
            mod = chain(((None, ()),), mod)
        for name, arguments in mod:
            try:
                arg = self.handler(arg, name, *arguments)
            except Exception as error:
                raise ArgumentTypeError(F'failed to apply modifier {name} to incoming data: {error}')
            if callable(arg):
                if data is None:
                    raise TooLazy
                arg = arg(data)
        return arg

    def handler(self, expression: str):
        """
        This method is overwritten by children of `refinery.lib.argformats.DelayedArgument`
        to implement the default handler.
        """
        raise NotImplementedError


class DelayedBinaryArgument(DelayedArgument):

    @DelayedArgumentDispatch
    def handler(self, expr: str) -> bytes:
        try:
            return open(expr, 'rb').read()
        except Exception:
            pass
        return utf8(expr)

    @handler.register('s', final=True)
    def s(self, string: str) -> bytes:
        """
        The final modifier `s:string` returns the UTF-8 encoded representation of `string`.
        """
        return string.encode('UTF8')

    @handler.register('u', final=True)
    def u(self, string: str) -> bytes:
        """
        The final modifier `u:string` returns the UTF16 (little endian without BOM) encoded
        representation of `string`.
        """
        return string.encode('UTF-16LE')

    @handler.register('a', final=True)
    def a(self, string: str) -> bytes:
        """
        The final modifier `a:string` returns the latin-1 encoded representation of `string`.
        """
        return string.encode('LATIN-1')

    @handler.register('H', 'h', final=True)
    def h(self, string: str) -> bytes:
        """
        The final modifier `h:string` (or `H:string`) returns the hex decoding of `string`.
        """
        import base64
        return base64.b16decode(string, casefold=True)

    @handler.register('f', 'file', final=True)
    def file(self, path: str) -> bytes:
        """
        The final modifier `f:path` or `file:path` returns the contents of the file located
        at the given path.
        """
        return open(path, 'rb').read()

    @handler.register('range', final=True)
    def range(self, region: str) -> bytes:
        """
        Implements the final modifier `range:bounds` to generate a sequence of bytes, where
        `bounds` is parsed as a `refinery.lib.argformats.sliceobj` with one exception: If
        `bounds` is just a single integer, it is interpreted as the upper bound for a sequence
        of bytes starting at zero.
        """
        try:
            bounds = number(region)
            return bytearray(range(bounds))
        except ValueError:
            pass
        bounds = sliceobj(region)
        if bounds.stop is None:
            raise ArgumentTypeError('cannot generate unbounded byte sequence.')
        return bytearray(range(bounds.start or 0, bounds.stop, bounds.step or 1))

    @handler.register('c', 'copy', final=True)
    def copy(self, region: str) -> bytes:
        """
        Implements the final modifier `c:region` or `copy:region`, where `region` is parsed
        as a `refinery.lib.argformats.sliceobj`. The result contains the corresponding slice
        of the input data.
        """
        bounds = sliceobj(region)
        return lambda d: d[bounds]

    @handler.register('x', 'cut', final=True)
    def cut(self, region: str) -> bytes:
        """
        `x:region` and `cut:region` work like `refinery.lib.argformats.DelayedBinaryArgument.copy`,
        but the corresponding bytes are also removed from the input data.
        """
        def extract(data: bytearray):
            result = data[bounds]
            data[bounds] = []
            return result
        bounds = sliceobj(region)
        return extract


def multibin(expression: Union[str, bytes, bytearray]) -> Union[bytes, DelayedBinaryArgument]:
    """
    This is the argument parser type that uses `refinery.lib.argformats.DelayedBinaryArgument`.
    """
    if not isinstance(expression, str):
        return bytes(expression)
    arg = DelayedBinaryArgument(expression)
    with suppress(TooLazy):
        return arg()
    return arg


class DelayedNumbinArgument(DelayedArgument):
    """
    A parser for sequences of numeric arguments. As `refinery.lib.argformats.DelayedNumbinArgument.handler`
    uses `refinery.lib.argformats.multibin`, it is possible to use any handler specified in
    `refinery.lib.argformats.DelayedBinaryArgument` as long as these handlers precede any of the handlers
    defined here.
    """

    _EV_PARSER = PythonExpression('N')

    def _mbin(self, expr: str) -> bytes:
        binary = multibin(expr)
        if not binary:
            raise ArgumentTypeError('received empty binary argument')
        return binary

    def _iter(self, unknown):
        if hasattr(unknown, '__iter__'):
            it = list(unknown)
            if all(isinstance(t, int) for t in it):
                return it
        if isinstance(unknown, int):
            return (unknown,)
        raise ArgumentTypeError(
            F'numbin parser encountered {unknown} of type {type(unknown).__name__}, '
            F'but only integers are supported.'
        )

    @DelayedArgumentDispatch
    def handler(self, expression: str) -> Iterable[int]:
        """
        The default handler: Attempts to parse the input expression as an integer and uses
        `refinery.lib.argformats.multibin` to parse it if that fails.
        """
        try:
            return (int(expression, 0),)
        except ValueError:
            return self._mbin(expression)

    @handler.register('ev', final=True)
    def ev(self, expression: str) -> Iterable[int]:
        """
        Final modifier `ev:expression`; uses a `refinery.lib.argformats.PythonExpression`
        parser to process expressions that may contain the variable `N` whose value will be
        the size of the input data.
        """
        ev = self._EV_PARSER(expression)
        try:
            return self._iter(ev())
        except Exception:
            return lambda d: self._iter(ev(N=len(d)))

    @handler.register('unpack', final=True)
    def unpack(self, expression: str) -> Iterable[int]:
        """
        Final modifier `unpack:[#]size:expression`; uses `refinery.lib.chunks.unpack` to
        convert a sequence of bytes into a sequence of numbers by unpacking them. The `expression`
        parameter is parsed with `refinery.lib.argformats.multibin` yielding this byte string.
        The `size` has to be an integer expression specifying the size of each encoded number in
        bytes. The optional hash tag modifier preceding the size indicates that the parser
        should use big endian rather than the default, little endian.
        """
        from .chunks import unpack
        size, expression = expression.split(':', 1)
        little_endian = True
        if size.startswith('#'):
            little_endian = False
            size = size[1:]
        try:
            size = int(size, 0)
        except ValueError:
            raise ArgumentTypeError(
                'the syntax is unpack:[!]size:bytes where size is an integer '
                'and bytes a multibin expression. You can specify the exclamation '
                'mark to use network (big endian) byte order.'
            )
        mbin = self._mbin(expression)
        if not callable(mbin):
            return unpack(mbin, size, little_endian)
        return lambda d: unpack(mbin(d), size, little_endian)

    @handler.register('inc')
    def inc(self, it: Iterable[int], wrap=None) -> Iterable[int]:
        """
        The modifier `inc:it` or `inc(wrap):it` expects a sequence `it` of integers
        (a binary string is interpreted as the sequence of its byte values), iterates it
        cyclically and perpetually adds an increasing counter to the result. If `wrap`
        is specified, then the counter is reduced modulo this number.
        """
        def delay(_):
            k = cycle(range(number(wrap))) if wrap else count()
            for item in cycle(it):
                yield item + next(k)
        return delay

    @handler.register('dec')
    def dec(self, it: Iterable[int], wrap=None) -> Iterable[int]:
        """
        Identical to `refinery.lib.argformats.DelayedNumbinArgument.inc`, but the counter
        is subtracted from `it`.
        """
        def delay(_):
            k = cycle(range(number(wrap))) if wrap else count()
            for item in cycle(it):
                yield item - next(k)
        return delay


class DelayedRegexpArgument(DelayedArgument):
    """
    A parser for regular expressions arguments.
    """

    @DelayedArgumentDispatch
    def handler(self, expression: str) -> bytes:
        """
        The default handler encodes the input expression as latin-1 to return a binary
        string regular expression.
        Furthermore, the use of named patterns from `refinery.lib.patterns.formats` and
        `refinery.lib.patterns.indicators` is possible by means of the extension format
        `(??name)`. For example, the pattern `e:((??url)\\x00){4}` will match a sequence
        of four URL strings which are all terminated with a null character.
        """
        if '(??' in expression:
            import re
            from .patterns import formats, indicators

            def replace(match):
                name = match.group(1)
                return '(?:{})'.format(formats.get(
                    name, indicators.get(name, match.group(0))))

            expression = re.sub(
                R'\(\?\?({}|{})\)'.format(
                    '|'.join(p.name for p in formats),
                    '|'.join(p.name for p in indicators)
                ),
                replace,
                expression
            )

        return expression.encode('latin-1')

    @handler.register('yara')
    def yara(self, pattern: bytes) -> bytes:
        """
        The handler `yara:pattern` converts YARA syntax wildcard hexadecimal expressions
        into regular expressions. For example, `D?` is translated to `[\\xD0-\\xDF]`, the
        expression `[2-6]` becomes `.{2,6}`, and `?D` becomes the following substring:
        ```
        [\\x0D\\x1D\\x2D\\x3D\\x4D\\x5D\\x6D\\x7D\\x8D\\x9D\\xAD\\xBD\\xCD\\xDD\\xED\\xFD]
        ```
        Only two-letter hexadecimal sequences with optional `?` wildcards and wildcard
        ranges such as `[2-6]` are substituted, all other characters in the pattern are
        left unchanged.
        """
        import re

        def y2r(match):
            expr = match.group(0)
            if expr == B'??':
                return B'.'
            if B'?' not in expr:
                return BR'\x%s' % expr
            if expr.endswith(B'?'):
                return BR'[\x%c0-\x%cF]' % (expr[0], expr[0])
            return BR'[%s]' % BR''.join(
                BR'\x%x%c' % (k, expr[1]) for k in range(0x10)
            )

        def yara_range(rng):
            return B'.{%s}' % B','.join(t.strip() for t in rng[1:-1].split(B'-'))

        pattern = re.split(BR'(\[\s*\d+(?:\s*-\s*\d+)?\s*\])', pattern)
        pattern[0::2] = [re.sub(BR'[A-F0-9?]{2}', y2r, c) for c in pattern[::2]]
        pattern[1::2] = [yara_range(b) for b in pattern[1::2]]
        return B''.join(pattern)

    @handler.register('escape')
    def escape(self, str: bytes) -> bytes:
        """
        The handler `escape:str` returns a regular expression which matches the exact
        string sequence given by `str`, with special regular expression control characters
        escaped.
        """
        import re
        return re.escape(str)


def numbin(expression: Union[int, str]) -> Union[Iterable[int], DelayedNumbinArgument]:
    """
    This is the argument parser type that uses `refinery.lib.argformats.DelayedNumbinArgument`.
    """
    if isinstance(expression, int):
        return (expression,)
    arg = DelayedNumbinArgument(expression)
    with suppress(TooLazy):
        return arg()
    return arg


def regexp(expression: str) -> Union[int, bytes, DelayedRegexpArgument]:
    """
    This is the argument parser type that uses `refinery.lib.argformats.DelayedRegexpArgument`.
    """
    arg = DelayedRegexpArgument(expression)
    with suppress(TooLazy):
        return arg()
    return arg


def OptionFactory(options: Mapping[str, Any]):
    """
    The factory produces an argument parser type that accepts the keys of `options`
    as possible values and causes the parsed argument to contain the corresponding
    value from the `options` dictionary.
    """

    class Option():
        def __init__(self, name: str):
            if name not in options:
                raise ValueError(
                    'the option %s is not one of these: %s' % (name, list(options)))
            self.mode = options[name]
            self.name = name

        def __eq__(self, other):
            return str(other) == self.name

        def __hash__(self):
            return hash(self.name)

        def __str__(self):
            return self.name

        def __repr__(self):
            return self.name

        @property
        def value(self):
            return self.mode

    return Option


def extract_options(symbols, prefix='MODE_'):
    """
    A helper function to extract all numeric constants from modules that have a certain
    prefix. `refinery.units.crypto.cipher.StandardCipherUnit` uses this to extract the
    block cipher modes of operation from block cipher modules of the `pycryptodome` library.
    """
    candidates = {
        k[len(prefix):]: getattr(symbols, k, None)
        for k in dir(symbols) if k.startswith(prefix)}
    return {k: v for k, v in candidates.items() if isinstance(v, int)}


def pending(argument: Union[Any, List[Any]]) -> bool:
    """
    This function returns a boolean value which indicates whether the given
    argument is a `refinery.lib.argformats.LazyEvaluation`.
    """
    if isinstance(argument, list):
        return any(pending(element) for element in argument)
    return isinstance(argument, LazyEvaluation)


def manifest(argument: Union[Any, List[Any]], data: bytearray) -> Union[Any, List[Any]]:
    """
    Returns the manifestation of a `refinery.lib.argformats.LazyEvaluation`
    on the given data. This function can change the data.
    """
    if isinstance(argument, list):
        return [manifest(x, data) for x in argument]
    return argument(data) if isinstance(argument, LazyEvaluation) else argument


def request(parser: Callable[[str], Any], predicate: Callable[[Any], bool], errmsg: str):
    """
    Turns an existing `parser` into one where the parsed data satisfies
    the given `predicate`, which is a callable returning a bool. If the
    parsed data is not pending and the predicate returns false, the
    exception `ArgumentTypeError(errmsg)` is generated.
    """
    @wraps(parser)
    def wrapped(s: str):
        result = parser(s)
        if not pending(result) and not predicate(result):
            raise ArgumentTypeError(errmsg)
        return result
    return wrapped



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import io

try:
    from Registry.Registry import Registry
except ModuleNotFoundError:
    Registry = None

from . import PathExtractorUnit


class winreg(PathExtractorUnit):
    """
    Extract values from a Windows registry hive.
    """
    def _walk(self, key, *path):
        here = '/'.join(path)
        if not self._check_reachable(here):
            return
        for value in key.values():
            vpath = F'{here}/{value.name()}'
            self.log_debug('crawl:', vpath)
            matching = self._check_path(vpath)
            if matching:
                self.log_info('match:', vpath)
                yield value.raw_data()
        for subkey in key.subkeys():
            yield from self._walk(subkey, *path, subkey.name())

    def process(self, data):
        with io.BytesIO(data) as stream:
            yield from self._walk(Registry(stream).root())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from itertools import cycle

from ... import Unit
from ....lib.decorators import unicoded


class vigenere(Unit):
    """
    Encryption and decryption using the Vigenère-Bellaso polyalphabetic cipher.
    """

    def interface(self, argp):
        argp.add_argument(
            '-c', '--case-sensitive',
            action='store_true',
            help='Unless this option is set, the key will be case insensitive and '
                 'the alphabet is assumed to contain only lowercase letters. Any '
                 'uppercase letter is transformed using the same shift as would be '
                 'the lowercase variant, but case is retained.'
        )
        argp.add_argument(
            '-i', '--ignore-unknown',
            action='store_true',
            help='Unless this option is set, the key stream will be iterated even '
                 'for letters that are not contained in the alphabet.'
        )
        argp.add_argument(
            'key', type=str, help='The encryption key')
        argp.add_argument(
            'alphabet',
            type=str,
            nargs='?',
            default='abcdefghijklmnopqrstuvwxyz',
            help='The alphabet which should be used, default is the Latin one: '
                 '"abcdefghijklmnopqrstuvwxyz"'
        )
        return super().interface(argp)

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        if not self.args.case_sensitive:
            self.args.key = self.args.key.lower()
        alphabet = ''
        for c in self.args.alphabet:
            if not self.args.case_sensitive:
                c = c.lower()
            if c not in alphabet:
                alphabet += c
        if alphabet != self.args.alphabet:
            self.log_warn(F'correcting alphabet to: {alphabet}')
            self.args.alphabet = alphabet
        if not set(self.args.key) <= set(self.args.alphabet):
            raise ValueError('key contains letters which are not from the given alphabet')

    def _tabula_recta(self, data, reverse=True):
        keystream = cycle(self.args.key)
        alphabet_size = len(self.args.alphabet)
        for letter in data:
            uppercase = not self.args.case_sensitive and letter.isupper()
            if uppercase:
                letter = letter.lower()
            try:
                position = self.args.alphabet.index(letter)
            except ValueError:
                yield letter
                if not self.args.ignore_unknown:
                    next(keystream)
                continue
            shift = self.args.alphabet.index(next(keystream))
            if reverse:
                position -= shift
            else:
                position += shift
            result = self.args.alphabet[position % alphabet_size]
            yield result.upper() if uppercase else result

    @unicoded
    def process(self, data):
        return ''.join(self._tabula_recta(data, True))

    @unicoded
    def reverse(self, data):
        return ''.join(self._tabula_recta(data, False))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import gzip

from .. import Unit


class gz(Unit):
    """
    GZip compression and decompression.
    """

    def process(self, data):
        return gzip.decompress(data)

    def reverse(self, data):
        return gzip.compress(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pefile

from . import HashUnit


class imphash(HashUnit):
    """
    Implements the import hash for PE files.
    """

    def process(self, data):
        th = pefile.PE(data=data).get_imphash()
        return th.encode(self.codec) if self.args.text else bytes.fromhex(th)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import json
try:
    import phpserialize as php
except ModuleNotFoundError:
    php = None

from .. import Unit


class dsphp(Unit):
    """
    Deserialize PHP serialized data and re-serialize as JSON.
    """

    class _encoder(json.JSONEncoder):
        def default(self, obj):
            try:
                return super().default(obj)
            except TypeError:
                pass
            if isinstance(obj, bytes) or isinstance(obj, bytearray):
                return obj.decode('utf8')
            if isinstance(obj, php.phpobject):
                return obj._asdict()

    def process(self, data):
        return json.dumps(
            php.loads(
                data,
                object_hook=php.phpobject,
                decode_strings=True
            ),
            indent=4,
            cls=self._encoder
        ).encode(self.codec)



#!/usr/bin/env python3
import re

from base64 import urlsafe_b64decode
from functools import wraps
from html import unescape
from urllib.parse import unquote, urlparse, parse_qs

from .. import Unit
from ...lib.decorators import unicoded


def unguard(pattern, flags=re.IGNORECASE):
    def decorator(method):
        @wraps(method)
        def _method(self, data):
            return re.sub(pattern, lambda m: method(self, m), data, flags=flags)
        return _method
    return decorator


class urlguards(Unit):
    """
    Restores the original URLs from their 'protected' versions as generated by
    Outlook protection and ProofPoint.
    """

    _PP3RLENC = {
        letter: rl for rl, letter in enumerate(
            'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
            'abcdefghijklmnopqrstuvwxyz'
            '0123456789-_', 2
        )
    }

    @unguard(r'https?://urldefense(?:\.proofpoint)?\.com/v([12])/url\?([:;/_=!?#&.,\w\%\-\+|]+)')
    def proofpointV2(self, match):
        version = int(match.group(1))
        self.log_info('proofpoint match:', version)
        argmatch = re.match(
            R'^u=(.+?)&(?:amp;)?{}='.format('k' if version == 1 else '[dc]'),
            match.group(2),
            flags=re.DOTALL
        )
        if not argmatch:
            self.log_warn('not able to translate unexpected proofpoint format:', match)
            return match.group(0)
        encoded = argmatch.group(1)
        if match.group(1) == '2':
            encoded = encoded.translate(str.maketrans('-_', '%/'))
        return unescape(unquote(encoded))

    @unguard(r'https?://urldefense(?:\.proofpoint)?\.com/v3/__(.+?)__;(.*?)![-\w!?$]+')
    def proofpointV3(self, match):
        data = unquote(match.group(1))
        cmap = match.group(2) + '=' * (-len(match.group(2)) % 4)
        cmap = urlsafe_b64decode(cmap).decode('UTF-8')
        cursor = 0
        result = ''
        for k in range(len(cmap)):
            ast = data.find('*', cursor)
            if ast < 0:
                break
            result += data[cursor:ast]
            if data[ast + 1] == '*':
                end = self._PP3RLENC[data[ast + 2]]
                result += cmap[k:end]
                ast += 2
            else:
                result += cmap[k]
            cursor = ast + 1
        self.log_debug(result)
        self.log_debug(data[cursor:])
        return result + data[cursor:]

    @unguard(r'https?://\w+.safelinks\.protection\.outlook\.com/([:;/_=!?#&.,\w\%\-\+|]+)')
    def outlook(self, match):
        result = match.group(0)
        self.log_info('outlook match:', result)
        parsed = urlparse(result)
        params = parse_qs(parsed.query)
        try:
            result = unquote(params['url'][0])
        except Exception:
            pass
        return result

    @unguard(r'https?://outlook.office.com/actions/ei\?u=([:;/_=!?#&.,\w\%\-\+|]+)')
    def outlook_image_proxy(self, match):
        return unquote(match.group(1))

    @unguard(r'https?://(?:[\w-]+\.)?trendmicro.com(?::\d+)?/wis/clicktime/v[12]/(?:query|clickthrough)[:;/_=!?#&.,\w\%\-\+|]+')
    def trendmicro(self, match):
        result = match.group(0)
        self.log_info('trendmicro match:', result)
        parsed = urlparse(result)
        params = parse_qs(parsed.query)
        try:
            result = unquote(params['url'][0])
        except Exception:
            pass
        return result

    @unicoded
    def process(self, data: str) -> str:
        newsize, size = 0, len(data)
        while newsize != size:
            for handler in (
                self.proofpointV2,
                self.proofpointV3,
                self.outlook,
                self.outlook_image_proxy,
                self.trendmicro
            ):
                data = handler(data)
            size = newsize
            newsize = len(data)
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re

from fnmatch import fnmatch
from ipaddress import ip_address
from urllib.parse import urlparse

from . import PatternExtractor
from .. import RefineryCriticalException
from ...lib.patterns import indicators


class xtp(PatternExtractor):
    """
    Extract Patterns: Uses regular expressions to extract indicators from the
    input data and optionally filters these results heuristically. The unit is
    designed to extract indicators such as domain names and IP addresses, see
    below for a complete list. To extract data formats such as hex-encoded
    data, use `refinery.carve`.
    """

    _LEGITIMATE_HOSTS = [
        'adobe.com',
        'aka.ms',
        'apache.org',
        'apple.com',
        'azure.com',
        'curl.haxx.se',
        'digicert.com',
        'globalsign.com',
        'globalsign.net',
        'iana.org',
        'live.com',
        'microsoft.com',
        'msdn.com',
        'msn.com',
        'office.com',
        'openssl.org',
        'openxmlformats.org',
        'python.org',
        'skype.com',
        'sway-cdn.com',
        'sway-extensions.com',
        'symantec.com',
        'symauth.com',
        'symcb.com',
        'thawte.com',
        'verisign.com',
        'w3.org',
        'xml.org',
        'xmlsoap.org',
        'yahoo.com',
    ]

    _DOMAIN_WHITELIST = [
        'system.net',
        'wscript.shell',
    ]

    def interface(self, argp):
        argp.add_argument('-f', '--filter', action='store_true', help=(
            'If this setting is enabled, the xtp unit will attempt to reduce the number '
            'of false positives by certain crude heuristics.'))
        argp.add_argument('pattern', metavar='PATTERN', type=str, nargs='*',
            default=['hostname', 'url', 'email'], help=(
                'Choose the pattern to extract, defaults are hostname, url, and email. '
                'Use an asterix character to select all available patterns. The available '
                'patterns are: {}'.format(', '.join(p.name for p in indicators))
            )
        )
        return super().interface(argp)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        patterns = {
            p for name in self.args.pattern for p in indicators if fnmatch(p.name, name)
        }
        if indicators.hostname in patterns:
            patterns.remove(indicators.hostname)
            patterns.add(indicators.ipv4)
            patterns.add(indicators.domain)
        patterns = [F'(?P<{p.name}>{p.value})' for p in patterns]
        if not patterns:
            raise RefineryCriticalException('The given mask does not match any known indicator pattern.')
        self.pattern = '|'.join(patterns)
        self.log_debug(F'using pattern: {self.pattern}')
        self.pattern = re.compile(self.pattern.encode(self.codec))

    @classmethod
    def _check_match(cls, data, pos, name, value):
        if name == 'ipv4':
            ocets = [int(x) for x in value.split(B'.')]
            if ocets.count(0) >= 3:
                return None
            for area in (
                data[pos - 20 : pos + 20],
                data[pos * 2 - 40 : pos * 2 + 40 : 2],
                data[pos * 2 - 41 : pos * 2 + 39 : 2]
            ):
                if B'version' in area.lower():
                    return None
            ip = ip_address(value.decode(cls.codec))
            if not ip.is_global and not ip.is_private:
                return None
        elif name in ('url', 'socket', 'domain'):
            ioc = value.decode(cls.codec)
            if '://' not in ioc: ioc = F'TCP://{ioc}'
            host = urlparse(ioc).netloc.split(':', 1)[0].lower()
            if any(host == w or host.endswith(F'.{w}') for w in cls._LEGITIMATE_HOSTS):
                return None
            if any(host == w for w in cls._DOMAIN_WHITELIST):
                return None
            if name == 'domain':
                hostparts = host.split('.')
                # These heuristics attempt to filter out member access to variables in
                # scripts which can be mistaken for domains because of the TLD inflation
                # we've had.
                if len(hostparts) == 2 and hostparts[0] == 'this':
                    return None
                if len(hostparts[-2]) < 3:
                    return None
                if any(x.startswith('_') for x in hostparts):
                    return None
                if len(hostparts[-1]) > 3 and len(set(re.findall(R'{}(?:\.\w+)+'.format(hostparts[0]).encode('ascii'), data))) > 2:
                    return None
        elif name == 'path':
            if len(value) < 8:
                return None
            if len(value) > 16 and len(re.findall(RB'\\x\d\d', value)) > len(value) // 10:
                return None
        return value

    def process(self, data):
        whitelist = set()

        def check(match):
            for name, value in match.groupdict().items():
                if value is not None:
                    break
            else:
                raise RefineryCriticalException('Received empty match.')
            if value in whitelist:
                return None
            result = self._check_match(data, match.start(), name, value)
            if result is not None:
                return result
            whitelist.add(value)

        transforms = None if not self.args.filter else [check]
        yield from self.matches_processed(data, self.pattern, transforms)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Implements several popular block and stream ciphers.
"""
from typing import Iterable, Any
from Crypto.Util.Padding import pad, unpad

from ... import Unit, Executable, RefineryCriticalException, RefineryPartialResult
from ....lib.argformats import multibin, OptionFactory, extract_options


class CipherExecutable(Executable):
    """
    A metaclass for the abstract class `refinery.units.crypto.cipher.CipherUnit` which
    normalizes the class variable `__key_sizes__` containing an iterable of all possible
    key sizes that are acceptable for the represented cipher.
    """
    def __init__(cls, name, bases, nmspc, abstract=False):

        if cls.__key_sizes__:
            try:
                iter(cls.__key_sizes__)
            except TypeError:
                cls.__key_sizes__ = (cls.__key_sizes__,)
        else:
            cls.__key_sizes__ = ()

        return super(CipherExecutable, cls).__init__(
            name, bases, nmspc, abstract=abstract
        )


class CipherUnit(Unit, metaclass=CipherExecutable, abstract=True):
    __blocksize__ = 1
    __key_sizes__ = None

    @property
    def __stream__(self) -> bool:
        return self.__blocksize__ == 1

    @property
    def maximum_key_size(self) -> int:
        return max(self.__key_sizes__)

    def interface(self, argp):

        if self.__blocksize__ > 1:
            argp.add_argument(
                '-b', '--block-align', action='store_true', help=(
                    F'Pad all input data with zero bytes to a multiple of the block '
                    F'size ({self.__blocksize__} bytes).'
                )
            )

            pdng = argp.add_mutually_exclusive_group()
            popt = ['PKCS7', 'ISO7816', 'X923', 'RAW']
            pdng.add_argument(
                '-P', '--padding', dest='padding', choices=popt,
                default=popt[:-1], metavar='ALG', nargs=1, help=(
                    F'Choose a padding algorithm ({", ".join(popt)}). The RAW algorithm does '
                    F'nothing. By default, all other algorithms are attempted. In most cases, '
                    F'the data was not correctly decrypted if none of these work.'
                )
            )

        argp.add_argument('key', type=multibin, help='encryption key')

        if not self.__stream__:
            argp.add_argument('--iv', type=multibin, default=bytes(self.__blocksize__), nargs='?',
                help='Specifies the initialization vector. If none is specified, then a block of zero bytes is used.')

        return super().interface(argp)

    def __init__(self, *args, **kw):
        if self.__key_sizes__:
            try:
                iter(self.__key_sizes__)
            except TypeError:
                self.__key_sizes__ = (self.__key_sizes__,)
        else:
            self.__key_sizes__ = ()
        super().__init__(*args, **kw)

    @property
    def key_size_chosen(self) -> int:
        try:
            return self.args.key_size
        except AttributeError:
            return self.maximum_key_size

    @property
    def key(self) -> bytes:
        return self.args.key

    @property
    def iv(self) -> bytes:
        if not self.__stream__ and len(self.args.iv) != self.__blocksize__:
            raise ValueError(
                F'The IV has length {len(self.args.iv)} but the block size '
                F'is {self.__blocksize__}.'
            )
        return self.args.iv

    def decrypt(self, data: bytes) -> bytes:
        raise NotImplementedError

    def encrypt(self, data: bytes) -> bytes:
        raise NotImplementedError

    def _zpad(self, data):
        if not self.__stream__ and self.args.block_align:
            overhang = len(data) % self.__blocksize__
            if overhang:
                padding = self.__blocksize__ - overhang
                self.log_info(F'appending {padding} padding null bytes')
                return data + B'\0' * padding
        return data

    def reverse(self, data: bytes) -> bytes:
        data = self._zpad(data)
        if not self.__stream__ and self.args.padding:
            padding = self.args.padding[0]
            if padding != 'RAW':
                self.log_info('padding method:', padding)
                data = pad(data, self.__blocksize__, padding.lower())
        data = self.encrypt(data)
        return data

    def process(self, data: bytes) -> bytes:
        return self.unpad(self.decrypt(self._zpad(data)))

    def unpad(self, data: bytes) -> bytes:
        if self.__stream__ or not self.args.padding:
            return data
        for p in self.args.padding:
            if p == 'RAW':
                return data
            try:
                return unpad(data, self.__blocksize__, p.lower())
            except Exception:
                pass
        else:
            raise RefineryPartialResult(
                'none of these paddings worked: ' + ', '.join(self.args.padding),
                partial=data
            )


class StandardCipherExecutable(CipherExecutable):

    def __new__(mcs, name, bases, nmspc, cipher=None):
        return super(StandardCipherExecutable, mcs).__new__(
            mcs, name, bases, nmspc, abstract=not cipher)

    def __init__(cls, name, bases, nmspc, cipher=None):
        abstract = not cipher
        if not abstract:
            cls.__cipher__ = cipher
            cls.__blocksize__ = cipher.block_size
            cls.__key_sizes__ = cipher.key_size
        else:
            cls.__cipher__ = NotImplemented
        return super(StandardCipherExecutable, cls).__init__(
            name, bases, nmspc, abstract=abstract)


class StandardCipherUnit(CipherUnit, metaclass=StandardCipherExecutable):

    def interface(self, argp):
        if not self.__stream__:
            modes = extract_options(self.__cipher__)
            if not modes:
                raise RefineryCriticalException(
                    F'cipher {self.__cipher__.__name__} is a block cipher module '
                    F'but no cipher block mode constants were found.'
                )
            argp.add_argument('mode', type=OptionFactory(modes), metavar='mode', choices=list(modes), help=(
                'Choose cipher mode to be used. Possible values are: {}.'.format(', '.join(modes))
            ))
        return super().interface(argp)

    def _get_cipher_instance(self, **optionals) -> Any:
        self.log_info(F'encryption key:', self.key.hex())
        if not self.__stream__:
            if not self.args.mode.name == 'ECB':
                optionals['IV'] = self.iv
                self.log_info('initial vector:', self.iv.hex())
            if self.args.mode:
                optionals['mode'] = self.args.mode.value
        try:
            return self.__cipher__.new(key=self.key, **optionals)
        except TypeError:
            # occurs when this mode requires no IV
            optionals.pop('IV')
            if self.iv:
                self.log_info('ignoring IV for mode', self.args.mode)
            return self.__cipher__.new(key=self.key, **optionals)

    def encrypt(self, data: bytes) -> bytes:
        return self._get_cipher_instance().encrypt(data)

    def decrypt(self, data: bytes) -> bytes:
        return self._get_cipher_instance().decrypt(data)


class StreamCipherUnit(CipherUnit, abstract=True):

    def __stream__(self) -> bool:
        return True

    def keystream(self) -> Iterable[bytes]:
        raise NotImplementedError

    def encrypt(self, data: bytes) -> bytes:
        key = iter(self.keystream())
        return bytes(b ^ next(key) for b in data)

    decrypt = encrypt


class LatinStreamCipher(StandardCipherUnit):

    def interface(self, argp):
        argp.add_argument('-N', '--nonce', default=B'REFINERY', type=multibin,
            help='Specify the one-time use nonce; the default value is "REFINERY".')
        return super().interface(argp)

    def _get_cipher_instance(self, **optionals) -> Any:
        self.log_info(F'one-time nonce:', self.args.nonce.hex())
        return super()._get_cipher_instance(nonce=self.args.nonce)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from Crypto.Protocol.KDF import PBKDF1 as PBKDF1_

from . import KeyDerivation


class PBKDF1(KeyDerivation):
    """PBKDF1 Key derivation"""

    _DEFAULT_ITER = None
    _DEFAULT_HASH = None
    _DEFAULT_SALT = bytes(8)

    def process(self, data):
        kwargs = dict(
            dkLen=self.args.size,
            count=self.iterations,
            hashAlgo=self.algorithm
        )
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        return PBKDF1_(data.decode(self.codec), self.salt, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from Crypto.Protocol.KDF import PBKDF2 as PBKDF2_

from . import KeyDerivation


class PBKDF2(KeyDerivation):
    """PBKDF2 Key derivation"""

    _DEFAULT_ITER = None
    _DEFAULT_SALT = None
    _DEFAULT_HASH = None

    def process(self, data):
        kwargs = dict(
            dkLen=self.args.size,
            count=self.args.iterations,
            hmac_hash_module=self.algorithm
        )
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        data = PBKDF2_(data.decode(self.codec), self.salt, **kwargs)
        return bytes(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re
import io

from zlib import crc32

from .. import Unit
from ...lib.argformats import number
from ...lib.decorators import unicoded


__all__ = [
    'Deobfuscator',
    'IterativeDeobfuscator'
]


class AutoDeobfuscationTimeout(RuntimeError):
    pass


def outside(*exceptions):
    """
    A decorator which allows to apply the transformation only to areas where
    a set of given regular expressions does not match. Here, this is mostly
    used to apply deobfuscations only to code outside of strings.
    """

    exclusion = '|'.join(F'(?:{e})' for e in exceptions)

    def excluded(method):
        def wrapper(self, data):
            with io.StringIO() as out:
                cursor = 0
                for m in re.finditer(exclusion, data, re.DOTALL):
                    out.write(method(self, data[cursor:m.start()]))
                    out.write(m.group(0))
                    cursor = m.end()
                out.write(method(self, data[cursor:]))
                return out.getvalue()
        return wrapper

    return excluded


class Deobfuscator(Unit, abstract=True):
    @unicoded
    def process(self, data: str) -> str:
        return self.deobfuscate(data)

    def deobfuscate(self, data: str) -> str:
        return data


class IterativeDeobfuscator(Deobfuscator, abstract=True):

    def interface(self, argp):
        argp.add_argument(
            '-t', '--timeout',
            type=number[2:],
            default=100,
            help='Specify the maximum number of iterations that may be '
                 'performed. The default is 100.'
        )
        return super().interface(argp)

    def process(self, data: bytes) -> bytes:
        previous = crc32(data)
        for _ in range(self.args.timeout):
            data = super().process(data)
            checksum = crc32(data)
            if checksum == previous:
                break
            previous = checksum
        else:
            raise AutoDeobfuscationTimeout()
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import json

from typing import Any
from enum import Enum
from hashlib import md5, sha1, sha256, sha512
from zlib import crc32

from .... import Unit
from ....encoding.hex import hex
from ....encoding.esc import esc
from ....encoding.url import url
from ....encoding.b64 import b64

from .....lib.dotnet.types import Blob


class BinaryEncoding(Enum):
    hex = hex
    esc = esc
    url = url
    b64 = b64


class BinaryDigest(Enum):
    md5 = md5
    crc32 = crc32
    sha1 = sha1
    sha256 = sha256
    sha512 = sha512


class JSONEncoderUnit(Unit, abstract=True):
    """
    An abstract unit that provides the interface for displaying parsed data
    as JSON. Since the parsed data may contain binary strings, a method for
    encoding binary strings needs to be specified, the default being hex
    encoding.
    """

    def interface(self, argp):
        enc = argp.add_mutually_exclusive_group()
        enc.add_argument(
            '-e', '--encoder', choices=[m.name for m in BinaryEncoding],
            default='hex', nargs='?', metavar='ENC', help=(
                'Select an encoder unit used to represent binary data in '
                'the JSON output. Available are: {}. The default encoder '
                'is hex.'.format(', '.join(m.name for m in BinaryEncoding))
            )
        )
        enc.add_argument(
            '-d', '--digest', choices=[d.name for d in BinaryDigest],
            default=None, nargs='?', metavar='HASH', help=(
                'Alternatively, select a hashing algorithm to digest binary data. '
                'The available algorithms are: {}.'.format(
                    ', '.join(d.name for d in BinaryDigest)
                )
            )
        )
        return super().interface(argp)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        class _encoder(json.JSONEncoder):
            unit = BinaryEncoding[self.args.encoder].value()
            digest = BinaryDigest[self.args.digest].value if self.args.digest else None

            def default(e, obj):
                try:
                    return super().default(obj)
                except TypeError:
                    pass

                if isinstance(obj, Blob):
                    obj = bytes(obj)
                if isinstance(obj, bytes) or isinstance(obj, bytearray):
                    if _encoder.digest:
                        return _encoder.digest(obj).hexdigest()
                    else:
                        return _encoder.unit.reverse(obj).decode('utf8')
                else:
                    return str(obj)

        self._encoder = _encoder

    def to_json(self, obj: Any) -> bytes:
        return json.dumps(obj, cls=self._encoder, indent=4).encode(self.codec)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from email.parser import BytesParser
from extract_msg.message import Message

from . import PathExtractorUnit


class EmailPart:
    def __init__(self, name, data):
        self.name = name
        self.data = data

    def __bool__(self):
        return bool(self.data)


class xtmail(PathExtractorUnit):
    """
    Extract files and body from EMail messages. The unit supports both the Outlook message format
    and regular MIME documents.
    """
    def _normalize_names(self, parts):
        names = set()
        unnamed = 0
        prefix = 'BODY'

        for part in parts:
            if not part:
                continue
            if part.name is None:
                unnamed += 1
            else:
                names.add(part.name)

        pw = len(str(unnamed))

        while {F'{prefix}{k:0{pw}}' for k in range(1, unnamed + 1)} & names:
            prefix = prefix + '_'

        for part in reversed(parts):
            if part.name is None:
                part.name = F'{prefix}{unnamed:0{pw}}'
                unnamed -= 1

    def _get_parts_outlook(self, data):
        def ensure_bytes(data):
            return data if isinstance(data, bytes) else data.encode(self.codec)
        with Message(data) as msg:
            parts = []
            if msg.body:
                parts.append(EmailPart(None, ensure_bytes(msg.body)))
            if msg.htmlBody:
                parts.append(EmailPart(None, ensure_bytes(msg.htmlBody)))
            for attachment in msg.attachments:
                parts.append(EmailPart(attachment.longFilename or attachment.shortFilename, attachment.data))
            return parts

    def _get_parts_regular(self, data):
        msg = BytesParser().parsebytes(data)
        return [EmailPart(part.get_filename(), part.get_payload(decode=True)) for part in msg.walk()]

    def process(self, data):
        try:
            parts = self._get_parts_outlook(data)
        except Exception as e:
            self.log_debug(F'failed parsing input as Outlook message: {e}')
            parts = self._get_parts_regular(data)

        self._normalize_names(parts)

        for part in parts:
            if part:
                if self._check_path(part.name):
                    self.log_info(part.name)
                    yield dict(data=part.data, path=part.name)
                else:
                    self.log_debug(part.name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import os.path

from itertools import cycle

from .. import Unit, RefineryCriticalException


class dump(Unit):
    """
    Dump incoming data to files on disk. It is possible to specify filenames with
    format fields using the `-f` or `--format` switch. The following format fields
    are supported:

         {ext} : Automatically guessed file extension
       {crc32} : CRC32 checksum of the data
       {index} : Index of the data in the input stream, starting at 1
      {length} : Size of the data in bytes
         {md5} : MD5 hash of the data
        {sha1} : SHA1 hash of the data
      {sha256} : SHA-256 hash of the data

    When not using formatted file names, the unit ingests as many incoming inputs
    as filenames were specified on the command line. Unless connected to a terminal,
    the remaining inputs will be forwarded on STDOUT. The `-t` or `--tee` switch
    can be used to forward all inputs, under all circumstances, regardless of
    whether or not they have been processed.

    If no file is specified, the first ingested input is dumped to the clipboard.
    """

    def interface(self, argp):
        argp.add_argument(
            'filenames',
            type=str,
            default=None,
            nargs='*',
            help='Output file',
            metavar='file'
        )
        argp.add_argument('-t', '--tee', action='store_true', 
            help='Forward all inputs to STDOUT.')

        mode = argp.add_mutually_exclusive_group()
        mode.add_argument('-s', '--stream', action='store_true', 
            help='Dump all incoming inputs to the same file.')

        mode.add_argument(
            '-m', '--meta', action='store_true', help=(
                'Automatically choose a filename. For multiple inputs, some '
                'units provide a file name in their output; in this case, '
                'this file name is used. Otherwise, the format {sha256} '
                'is used as the file name.'
            )
        )
        mode.add_argument(
            '-f', '--format', action='store_true', help=(
                'Provide format strings instead of a list of filenames. '
                'The format strings will be used cyclically to generate '
                'a file name for each input to be dumped.'
            )
        )

        return super().interface(argp)

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._reset()

    def _auto_extension(self, data, default='bin'):
        import re
        from ...lib.magic import magic, magicparse

        if not magic:
            self.log_warn(F'magic library not found, auto extension defaults to {default}')
            return default

        if not isinstance(data, bytes):
            data = bytes(data)

        mime = magicparse(data, mime=True)
        mime = mime.split(';')[0].lower()
        mtype, mext = mime.split('/')

        if mext == 'x-dosexec':
            description = magicparse(data)
            if re.search('executable', description):
                return 'dll' if '(DLL)' in description else 'exe'
        try:
            return {
                'octet-stream'                : 'bin',
                'plain'                       : 'txt',
                'javascript'                  : 'js',
                'java-archive'                : 'jar',
                'svg+xml'                     : 'svg',
                'x-icon'                      : 'ico',
                'wave'                        : 'wav',
                'x-pn-wav'                    : 'wav',
                'x-ms-wim'                    : 'wim',
                'vnd.android.package-archive' : 'apk',
                'vnd.ms-cab-compressed'       : 'cab',
                'x-apple-diskimage'           : 'dmg',
            }[mext]
        except KeyError:
            if 'gzip' in mext:
                import gzip
                ungz = gzip.decompress(data)
                ext1 = self._auto_extension(ungz)
                return ext1 + '.gz'
            xtype_match = re.match(
                r'^x-(\w{2,4})(-compressed)?$',
                mext,
                re.IGNORECASE
            )
            if xtype_match:
                return xtype_match.group(1)
            if len(mext) < 6 and re.match('[a-z]+', mext):
                return mext
            return default

    def _reset(self):
        self.stream = None
        self.exhausted = False
        if self.args.meta:
            self.iter_filenames = iter(())
        elif self.args.format:
            self.iter_filenames = cycle(self.args.filenames)
        else:
            self.iter_filenames = iter(self.args.filenames)
        if self.args.stream and len(self.args.filenames) > 1:
            raise ValueError('can only use one file in stream mode.')

    @property
    def _paste(self):
        return not bool(self.args.filenames) and not bool(self.args.format) and not self.args.meta

    def _open(self, filename, unc=False):
        filename = os.path.abspath(filename)
        base = os.path.dirname(filename)
        if not unc:
            self.log_info(filename)
        try:
            os.makedirs(base, exist_ok=True)
        except FileNotFoundError:
            if not unc and os.name == 'nt':
                filename = F'\\\\?\\{filename}'
                return self._open(filename, unc=True)
            raise
        else:
            return open(filename, 'wb')

    def _close(self):
        if self.stream:
            self.stream.flush()
            self.stream.close()
            self.stream = None

    def _format(self, filename, data, index=0):
        class DelayedFormatter(dict):
            def __missing__(_, key):
                if key == 'crc32':
                    from zlib import crc32
                    return F'{crc32(data) & 0xFFFFFFFF:08X}'
                if key == 'ext':
                    return self._auto_extension(data)
                if key in ('md5', 'sha1', 'sha256'):
                    import hashlib
                    algorithm = getattr(hashlib, key)
                    return algorithm(data).hexdigest()
                return '{' + key + '}'
        return filename.format_map(
            DelayedFormatter(dict(index=index, length=len(data)))
        )

    def process(self, data):
        if not self.exhausted:
            if self._paste:
                import codecs
                import pyperclip
                pyperclip.copy(codecs.decode(
                    data, 'utf-8', errors='backslashreplace'))
            elif not self.stream:
                # This should happen only when the unit is called from Python code
                # rather than via the command line.
                try:
                    filename = next(self.iter_filenames)
                except StopIteration:
                    raise RefineryCriticalException('the list of filenames was exhausted.')
                else:
                    with self._open(filename) as stream:
                        stream.write(data)
            else:
                self.stream.write(data)
                self.log_debug(F'wrote 0x{len(data):08X} bytes')
                if not self.args.stream:
                    self._close()
            forward_input_data = self.args.tee
        else:
            forward_input_data = self.args.tee or not self.isatty
            if not forward_input_data:
                self.log_debug(F'discarding unprocessed chunk of size {len(data)}.')

        if forward_input_data:
            yield data

    def filter(self, chunks):
        if self.exhausted:
            self._reset()

        if self._paste:
            it = iter(chunks)
            yield next(it)
            self.exhausted = True
            yield from it
            return

        for index, chunk in enumerate(chunks):
            if not self.exhausted:
                if not self.args.stream or not self.stream:
                    try:
                        if not self.args.meta:
                            filename = next(self.iter_filenames)
                        else:
                            filename = chunk['path']
                            if not filename:
                                import hashlib
                                filename = hashlib.sha256(chunk).hexdigest()
                    except StopIteration:
                        self.exhausted = True
                    else:
                        if self.args.format:
                            filename = self._format(filename, chunk, index + 1)
                        self.stream = self._open(filename)
            yield chunk

        self._close()
        self.exhausted = True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from Crypto.Cipher import AES

from . import StandardCipherUnit


class aes(StandardCipherUnit, cipher=AES):
    """
    AES encryption and decryption.
    """
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import re
import string

from ....lib.patterns import formats
from .. import Deobfuscator
from . import string_escape, string_quote


class deob_ps1_typecast(Deobfuscator):
    """
    Replaces sequences like [Char]120 to their string representation, in this
    case the string "x".
    """

    def deobfuscate(self, data):

        data = re.sub(
            R'\[(?:string|char\[\])\]\s*(%s)' % formats.ps1str,
            R'\1',
            data,
            flags=re.IGNORECASE
        )

        def char_literal(match):
            return string_quote(string_escape(chr(int(match.group(1), 0))))

        data = re.sub(
            R'\[char\]\s*0*(0x[0-9a-f]+|\d+)',
            char_literal,
            data,
            flags=re.IGNORECASE
        )

        def char_array(match):
            result = bytes(int(x, 0) for x in match.group(1).split(','))
            try:
                result = result.decode('ascii')
                if not all(x in string.printable for x in result):
                    raise ValueError
            except ValueError:
                return match.group(0)
            else:
                return string_quote(string_escape(result))

        data = re.sub(
            R'\s*'.join([
                R'\[char\[\]\]',
                R'\((',
                R'(?:\s*(?:0x[0-9a-f]+|\d+)\s*,)+',
                R'(?:0x[0-9a-f]+|\d+)',
                R')\)'
            ]),
            char_array,
            data,
            flags=re.IGNORECASE
        )

        return data
